from __future__ import annotations

import os
from typing import Literal, Mapping, Sequence

import slugify
from bec_lib import bec_logger
from qtpy.QtCore import Signal
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QInputDialog,
    QMessageBox,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

import bec_widgets.widgets.containers.qt_ads as QtAds
from bec_widgets import BECWidget, SafeProperty, SafeSlot
from bec_widgets.applications.views.view import ViewTourSteps
from bec_widgets.cli.rpc.rpc_widget_handler import widget_handler
from bec_widgets.utils import BECDispatcher
from bec_widgets.utils.colors import apply_theme
from bec_widgets.utils.rpc_decorator import rpc_timeout
from bec_widgets.utils.toolbars.actions import (
    ExpandableMenuAction,
    MaterialIconAction,
    WidgetAction,
)
from bec_widgets.utils.toolbars.bundles import ToolbarBundle
from bec_widgets.utils.toolbars.toolbar import ModularToolBar
from bec_widgets.utils.widget_state_manager import WidgetStateManager
from bec_widgets.widgets.containers.dock_area.basic_dock_area import DockAreaWidget
from bec_widgets.widgets.containers.dock_area.profile_utils import (
    SETTINGS_KEYS,
    default_profile_candidates,
    delete_profile_files,
    get_last_profile,
    is_profile_read_only,
    is_quick_select,
    list_profiles,
    list_quick_profiles,
    load_default_profile_screenshot,
    load_user_profile_screenshot,
    now_iso_utc,
    open_default_settings,
    open_user_settings,
    profile_origin,
    profile_origin_display,
    read_manifest,
    restore_user_from_default,
    set_last_profile,
    set_quick_select,
    user_profile_candidates,
    write_manifest,
)
from bec_widgets.widgets.containers.dock_area.settings.dialogs import (
    RestoreProfileDialog,
    SaveProfileDialog,
)
from bec_widgets.widgets.containers.dock_area.settings.workspace_manager import WorkSpaceManager
from bec_widgets.widgets.containers.dock_area.toolbar_components.workspace_actions import (
    WorkspaceConnection,
    workspace_bundle,
)
from bec_widgets.widgets.containers.main_window.main_window import BECMainWindowNoRPC
from bec_widgets.widgets.containers.qt_ads import CDockWidget
from bec_widgets.widgets.control.device_control.positioner_box import PositionerBox, PositionerBox2D
from bec_widgets.widgets.control.scan_control import ScanControl
from bec_widgets.widgets.editors.web_console.web_console import BECShell, WebConsole
from bec_widgets.widgets.plots.heatmap.heatmap import Heatmap
from bec_widgets.widgets.plots.image.image import Image
from bec_widgets.widgets.plots.motor_map.motor_map import MotorMap
from bec_widgets.widgets.plots.multi_waveform.multi_waveform import MultiWaveform
from bec_widgets.widgets.plots.scatter_waveform.scatter_waveform import ScatterWaveform
from bec_widgets.widgets.plots.waveform.waveform import Waveform
from bec_widgets.widgets.progress.ring_progress_bar.ring_progress_bar import RingProgressBar
from bec_widgets.widgets.services.bec_queue.bec_queue import BECQueue
from bec_widgets.widgets.services.bec_status_box.bec_status_box import BECStatusBox
from bec_widgets.widgets.utility.logpanel import LogPanel
from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton

logger = bec_logger.logger

_PROFILE_NAMESPACE_UNSET = object()

PROFILE_STATE_KEYS = {key: SETTINGS_KEYS[key] for key in ("geom", "state", "ads_state")}
StartupProfile = Literal["restore", "skip"] | str | None


class BECDockArea(DockAreaWidget):
    RPC = True
    PLUGIN = False
    USER_ACCESS = [
        "new",
        "widget_map",
        "widget_list",
        "workspace_is_locked",
        "attach_all",
        "delete_all",
        "delete",
        "set_layout_ratios",
        "describe_layout",
        "mode",
        "mode.setter",
        "list_profiles",
        "save_profile",
        "load_profile",
        "delete_profile",
    ]

    # Define a signal for mode changes
    mode_changed = Signal(str)
    profile_changed = Signal(str)

    def __init__(
        self,
        parent=None,
        mode: Literal["plot", "device", "utils", "user", "creator"] = "creator",
        default_add_direction: Literal["left", "right", "top", "bottom"] = "right",
        profile_namespace: str | None = None,
        auto_profile_namespace: bool = True,
        instance_id: str | None = None,
        auto_save_upon_exit: bool = True,
        enable_profile_management: bool = True,
        startup_profile: StartupProfile = "restore",
        **kwargs,
    ):
        self._profile_namespace_hint = profile_namespace
        self._profile_namespace_auto = auto_profile_namespace
        self._profile_namespace_resolved: str | None | object = _PROFILE_NAMESPACE_UNSET
        self._instance_id = slugify.slugify(instance_id, separator="_") if instance_id else None
        self._auto_save_upon_exit = auto_save_upon_exit
        self._profile_management_enabled = enable_profile_management
        self._startup_profile = self._normalize_startup_profile(startup_profile)
        super().__init__(
            parent, default_add_direction=default_add_direction, title="BEC Dock Area", **kwargs
        )

        # Initialize mode property first (before toolbar setup)
        self._mode = mode

        # Toolbar
        self.dark_mode_button = DarkModeButton(parent=self, toolbar=True)
        self.dark_mode_button.setVisible(enable_profile_management)
        self._setup_toolbar()
        self._hook_toolbar()

        # Popups
        self.save_dialog = None
        self.manage_dialog = None

        # Place toolbar above the dock manager provided by the base class
        self._root_layout.insertWidget(0, self.toolbar)

        # Populate and hook the workspace combo
        self._current_profile_name = None
        self._empty_profile_active = False
        self._empty_profile_consumed = False
        self._pending_autosave_skip: tuple[str, str] | None = None
        self._exit_snapshot_written = False
        self._refresh_workspace_list()

        # State manager
        self.state_manager = WidgetStateManager(
            self, serialize_from_root=True, root_id="BECDockArea"
        )

        # Developer mode state
        self._editable = None
        # Initialize default editable state based on current lock
        self._set_editable(True)  # default to editable; will sync toolbar toggle below

        # Apply the requested mode after everything is set up
        self.mode = mode
        self._fetch_initial_profile()

    @staticmethod
    def _normalize_startup_profile(startup_profile: StartupProfile) -> StartupProfile:
        """
        Normalize startup profile values.
        """
        if startup_profile == "":
            return None
        return startup_profile

    def _resolve_restore_startup_profile(self) -> str | None:
        """
        Resolve the profile name when startup profile is set to "restore".
        """
        combo = self.toolbar.components.get_action("workspace_combo").widget
        namespace = self.profile_namespace

        instance_id = self._last_profile_instance_id()
        if instance_id:
            inst_profile = get_last_profile(
                namespace=namespace, instance=instance_id, allow_namespace_fallback=False
            )
            if inst_profile and self._profile_exists(inst_profile, namespace):
                return inst_profile

        last = get_last_profile(namespace=namespace)
        if last and self._profile_exists(last, namespace):
            return last

        combo_text = combo.currentText().strip()
        if combo_text and self._profile_exists(combo_text, namespace):
            return combo_text

        return None

    def _fetch_initial_profile(self):
        startup_profile = self._startup_profile

        if startup_profile == "skip":
            logger.debug("Skipping startup profile initialization.")
            return

        if startup_profile == "restore":
            restored = self._resolve_restore_startup_profile()
            if restored:
                self._load_initial_profile(restored)
                return
            self._start_empty_workspace()
            return

        if startup_profile is None:
            self._start_empty_workspace()
            return

        self._load_initial_profile(startup_profile)

    def _load_initial_profile(self, name: str) -> None:
        """Load the initial profile."""
        self.load_profile(name)
        combo = self.toolbar.components.get_action("workspace_combo").widget
        combo.blockSignals(True)
        if not self._empty_profile_active:
            combo.setCurrentText(name)
        combo.blockSignals(False)

    def _start_empty_workspace(self) -> None:
        """
        Initialize the dock area in transient empty-profile mode.
        """
        if (
            getattr(self, "_current_profile_name", None) is None
            and not self._empty_profile_consumed
        ):
            self.delete_all()
            self._enter_empty_profile_state()

    def _customize_dock(self, dock: CDockWidget, widget: QWidget) -> None:
        prefs = getattr(dock, "_dock_preferences", {}) or {}
        if prefs.get("show_settings_action") is None:
            prefs = dict(prefs)
            prefs["show_settings_action"] = True
            dock._dock_preferences = prefs
        super()._customize_dock(dock, widget)

    @SafeSlot(popup_error=True)
    def new(
        self,
        widget: QWidget | str,
        *,
        closable: bool = True,
        floatable: bool = True,
        movable: bool = True,
        start_floating: bool = False,
        where: Literal["left", "right", "top", "bottom"] | None = None,
        tab_with: CDockWidget | QWidget | str | None = None,
        relative_to: CDockWidget | QWidget | str | None = None,
        show_title_bar: bool | None = None,
        title_buttons: Mapping[str, bool] | Sequence[str] | str | None = None,
        show_settings_action: bool | None = None,
        promote_central: bool = False,
        object_name: str | None = None,
        **widget_kwargs,
    ) -> QWidget | BECWidget:
        """
        Create a new widget (or reuse an instance) and add it as a dock.

        Args:
            widget(QWidget | str): Instance or registered widget type string.
            closable(bool): Whether the dock is closable.
            floatable(bool): Whether the dock is floatable.
            movable(bool): Whether the dock is movable.
            start_floating(bool): Whether to start the dock floating.
            where(Literal["left", "right", "top", "bottom"] | None): Dock placement hint relative to the dock area (ignored when
                ``relative_to`` is provided without an explicit value).
            tab_with(CDockWidget | QWidget | str | None): Existing dock (or widget/name) to tab the new dock alongside.
            relative_to(CDockWidget | QWidget | str | None): Existing dock (or widget/name) used as the positional anchor.
                When supplied and ``where`` is ``None``, the new dock inherits the
                anchor's current dock area.
            show_title_bar(bool | None): Explicitly show or hide the dock area's title bar.
            title_buttons(Mapping[str, bool] | Sequence[str] | str | None): Mapping or iterable describing which title bar buttons should
                remain visible. Provide a mapping of button names (``"float"``,
                ``"close"``, ``"menu"``, ``"auto_hide"``, ``"minimize"``) to booleans,
                or a sequence of button names to hide.
            show_settings_action(bool | None): Control whether a dock settings/property action should
                be installed. Defaults to ``False`` for the basic dock area; subclasses
                such as `BECDockArea` override the default to ``True``.
            promote_central(bool): When True, promote the created dock to be the dock manager's
                central widget (useful for editor stacks or other root content).
            object_name(str | None): Optional object name to assign to the created widget.
            **widget_kwargs: Additional keyword arguments passed to the widget constructor
                when creating by type name.

        Returns:
            BECWidget: The created or reused widget instance.
        """
        if show_settings_action is None:
            show_settings_action = True
        return super().new(
            widget,
            closable=closable,
            floatable=floatable,
            movable=movable,
            start_floating=start_floating,
            where=where,
            tab_with=tab_with,
            relative_to=relative_to,
            show_title_bar=show_title_bar,
            title_buttons=title_buttons,
            show_settings_action=show_settings_action,
            promote_central=promote_central,
            object_name=object_name,
            **widget_kwargs,
        )

    def _apply_dock_lock(self, locked: bool) -> None:
        if locked:
            self.dock_manager.lockDockWidgetFeaturesGlobally()
        else:
            self.dock_manager.lockDockWidgetFeaturesGlobally(QtAds.CDockWidget.NoDockWidgetFeatures)

    ################################################################################
    # Toolbar Setup
    ################################################################################

    def _setup_toolbar(self):
        self.toolbar = ModularToolBar(parent=self)

        plot_actions = {
            "waveform": (Waveform.ICON_NAME, "Add Waveform", "Waveform"),
            "scatter_waveform": (
                ScatterWaveform.ICON_NAME,
                "Add Scatter Waveform",
                "ScatterWaveform",
            ),
            "multi_waveform": (MultiWaveform.ICON_NAME, "Add Multi Waveform", "MultiWaveform"),
            "image": (Image.ICON_NAME, "Add Image", "Image"),
            "motor_map": (MotorMap.ICON_NAME, "Add Motor Map", "MotorMap"),
            "heatmap": (Heatmap.ICON_NAME, "Add Heatmap", "Heatmap"),
        }
        device_actions = {
            "scan_control": (ScanControl.ICON_NAME, "Add Scan Control", "ScanControl"),
            "positioner_box": (PositionerBox.ICON_NAME, "Add Device Box", "PositionerBox"),
            "positioner_box_2D": (
                PositionerBox2D.ICON_NAME,
                "Add Device 2D Box",
                "PositionerBox2D",
            ),
        }
        util_actions = {
            "queue": (BECQueue.ICON_NAME, "Add Scan Queue", "BECQueue"),
            "status": (BECStatusBox.ICON_NAME, "Add BEC Status Box", "BECStatusBox"),
            "progress_bar": (
                RingProgressBar.ICON_NAME,
                "Add Circular ProgressBar",
                "RingProgressBar",
            ),
            "terminal": (WebConsole.ICON_NAME, "Add Terminal", "WebConsole"),
            "bec_shell": (BECShell.ICON_NAME, "Add BEC Shell", "BECShell"),
            "log_panel": (LogPanel.ICON_NAME, "Add LogPanel - Disabled", "LogPanel"),
            "sbb_monitor": ("train", "Add SBB Monitor", "SBBMonitor"),
        }

        # Create expandable menu actions (original behavior)
        def _build_menu(key: str, label: str, mapping: dict[str, tuple[str, str, str]]):
            self.toolbar.components.add_safe(
                key,
                ExpandableMenuAction(
                    label=label,
                    actions={
                        k: MaterialIconAction(
                            icon_name=v[0], tooltip=v[1], filled=True, parent=self
                        )
                        for k, v in mapping.items()
                    },
                ),
            )
            b = ToolbarBundle(key, self.toolbar.components)
            b.add_action(key)
            self.toolbar.add_bundle(b)

        _build_menu("menu_plots", "Add Plot ", plot_actions)
        _build_menu("menu_devices", "Add Device Control ", device_actions)
        _build_menu("menu_utils", "Add Utils ", util_actions)

        # Create flat toolbar bundles for each widget type
        def _build_flat_bundles(category: str, mapping: dict[str, tuple[str, str, str]]):
            bundle = ToolbarBundle(f"flat_{category}", self.toolbar.components)

            for action_id, (icon_name, tooltip, widget_type) in mapping.items():
                # Create individual action for each widget type
                flat_action_id = f"flat_{action_id}"
                self.toolbar.components.add_safe(
                    flat_action_id,
                    MaterialIconAction(
                        icon_name=icon_name,
                        tooltip=tooltip,
                        filled=True,
                        parent=self,
                        label_text=widget_type,
                        text_position="under",
                    ),
                )
                bundle.add_action(flat_action_id)

            self.toolbar.add_bundle(bundle)

        _build_flat_bundles("plots", plot_actions)
        _build_flat_bundles("devices", device_actions)
        _build_flat_bundles("utils", util_actions)

        # Workspace
        spacer_bundle = ToolbarBundle("spacer_bundle", self.toolbar.components)
        spacer = QWidget(parent=self.toolbar.components.toolbar)
        spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.toolbar.components.add_safe("spacer", WidgetAction(widget=spacer, adjust_size=False))
        spacer_bundle.add_action("spacer")
        self.toolbar.add_bundle(spacer_bundle)

        self.toolbar.add_bundle(
            workspace_bundle(self.toolbar.components, enable_tools=self._profile_management_enabled)
        )
        self.toolbar.connect_bundle(
            "workspace", WorkspaceConnection(components=self.toolbar.components, target_widget=self)
        )

        # Dock actions
        self.toolbar.components.add_safe(
            "attach_all",
            MaterialIconAction(
                icon_name="zoom_in_map", tooltip="Attach all floating docks", parent=self
            ),
        )
        self.toolbar.components.get_action("attach_all").action.setVisible(
            self._profile_management_enabled
        )
        self.toolbar.components.add_safe(
            "screenshot",
            MaterialIconAction(icon_name="photo_camera", tooltip="Take Screenshot", parent=self),
        )
        self.toolbar.components.get_action("screenshot").action.setVisible(
            self._profile_management_enabled
        )
        dark_mode_action = WidgetAction(
            widget=self.dark_mode_button, adjust_size=False, parent=self
        )
        dark_mode_action.widget.setVisible(self._profile_management_enabled)
        self.toolbar.components.add_safe("dark_mode", dark_mode_action)

        bda = ToolbarBundle("dock_actions", self.toolbar.components)
        bda.add_action("attach_all")
        bda.add_action("screenshot")
        bda.add_action("dark_mode")
        self.toolbar.add_bundle(bda)

        self._apply_toolbar_layout()

        # Store mappings on self for use in _hook_toolbar
        self._ACTION_MAPPINGS = {
            "menu_plots": plot_actions,
            "menu_devices": device_actions,
            "menu_utils": util_actions,
        }

    def _hook_toolbar(self):
        def _connect_menu(menu_key: str):
            menu = self.toolbar.components.get_action(menu_key)
            mapping = self._ACTION_MAPPINGS[menu_key]

            # first two items not needed for this part
            for key, (_, _, widget_type) in mapping.items():
                act = menu.actions[key].action
                if widget_type == "LogPanel":
                    act.setEnabled(False)  # keep disabled per issue #644
                elif key == "terminal":
                    act.triggered.connect(
                        lambda _, t=widget_type: self.new(widget=t, closable=True, startup_cmd=None)
                    )
                elif key == "bec_shell":
                    act.triggered.connect(
                        lambda _, t=widget_type: self.new(
                            widget=t, closable=True, show_settings_action=False
                        )
                    )
                else:
                    act.triggered.connect(lambda _, t=widget_type: self.new(widget=t))

        _connect_menu("menu_plots")
        _connect_menu("menu_devices")
        _connect_menu("menu_utils")

        def _connect_flat_actions(mapping: dict[str, tuple[str, str, str]]):
            for action_id, (_, _, widget_type) in mapping.items():
                flat_action_id = f"flat_{action_id}"
                flat_action = self.toolbar.components.get_action(flat_action_id).action
                if widget_type == "LogPanel":
                    flat_action.setEnabled(False)  # keep disabled per issue #644
                else:
                    flat_action.triggered.connect(lambda _, t=widget_type: self.new(t))

        _connect_flat_actions(self._ACTION_MAPPINGS["menu_plots"])
        _connect_flat_actions(self._ACTION_MAPPINGS["menu_devices"])
        _connect_flat_actions(self._ACTION_MAPPINGS["menu_utils"])

        self.toolbar.components.get_action("attach_all").action.triggered.connect(self.attach_all)
        self.toolbar.components.get_action("screenshot").action.triggered.connect(self.screenshot)

    def _set_editable(self, editable: bool) -> None:
        self.workspace_is_locked = not editable
        self._editable = editable

        if self._profile_management_enabled:
            self.toolbar.components.get_action("attach_all").action.setVisible(editable)

    def _on_developer_mode_toggled(self, checked: bool) -> None:
        """Handle developer mode checkbox toggle."""
        self._set_editable(checked)

    ################################################################################
    # Workspace Management
    ################################################################################
    @SafeProperty(bool)
    def workspace_is_locked(self) -> bool:
        """
        Get or set the lock state of the workspace.

        Returns:
            bool: True if the workspace is locked, False otherwise.
        """
        return self._locked

    @workspace_is_locked.setter
    def workspace_is_locked(self, value: bool):
        """
        Set the lock state of the workspace. Docks remain resizable, but are not movable or closable.

        Args:
            value (bool): True to lock the workspace, False to unlock it.
        """
        self._locked = value
        self._apply_dock_lock(value)
        if self._profile_management_enabled:
            self.toolbar.components.get_action("save_workspace").action.setVisible(not value)
        for dock in self.dock_list():
            dock.setting_action.setVisible(not value)

    def _last_profile_instance_id(self) -> str | None:
        """
        Identifier used to scope the last-profile entry for this dock area.

        When unset, profiles are scoped only by namespace.
        """
        return self._instance_id

    def _resolve_profile_namespace(self) -> str | None:
        if self._profile_namespace_resolved is not _PROFILE_NAMESPACE_UNSET:
            return self._profile_namespace_resolved  # type: ignore[return-value]

        candidate = self._profile_namespace_hint
        if self._profile_namespace_auto:
            if not candidate:
                obj_name = self.objectName()
                candidate = obj_name if obj_name else None
            if not candidate:
                title = self.windowTitle()
                candidate = title if title and title.strip() else None
            if not candidate:
                mode_name = getattr(self, "_mode", None) or "creator"
                candidate = f"{mode_name}_workspace"
            if not candidate:
                candidate = self.__class__.__name__

        resolved = slugify.slugify(candidate, separator="_") if candidate else None
        if not resolved:
            resolved = "general"
        self._profile_namespace_resolved = resolved  # type: ignore[assignment]
        return resolved

    @property
    def profile_namespace(self) -> str | None:
        """Namespace used to scope user/default profile files for this dock area."""
        return self._resolve_profile_namespace()

    def _profile_exists(self, name: str, namespace: str | None) -> bool:
        return any(
            os.path.exists(path) for path in user_profile_candidates(name, namespace)
        ) or any(os.path.exists(path) for path in default_profile_candidates(name, namespace))

    def _write_snapshot_to_settings(self, settings, save_preview: bool = True) -> None:
        """
        Write the current workspace snapshot to the provided settings object.

        Args:
            settings(QSettings): The settings object to write to.
            save_preview(bool): Whether to save a screenshot preview.
        """
        self.save_to_settings(settings, keys=PROFILE_STATE_KEYS)
        self.state_manager.save_state(settings=settings)
        write_manifest(settings, self.dock_list())
        if save_preview:
            ba = self.screenshot_bytes()
            if ba and len(ba) > 0:
                settings.setValue(SETTINGS_KEYS["screenshot"], ba)
                settings.setValue(SETTINGS_KEYS["screenshot_at"], now_iso_utc())

        logger.info(f"Workspace snapshot written to settings: {settings.fileName()}")

    def _write_profile_settings(
        self,
        name: str,
        namespace: str | None,
        *,
        write_default: bool = True,
        write_user: bool = True,
        save_preview: bool = True,
    ) -> None:
        """
        Write profile settings to default and/or user settings files.

        Args:
            name: The profile name.
            namespace: The profile namespace.
            write_default: Whether to write to the default settings file.
            write_user: Whether to write to the user settings file.
            save_preview: Whether to save a screenshot preview.
        """
        if write_default:
            ds = open_default_settings(name, namespace=namespace)
            self._write_snapshot_to_settings(ds, save_preview=save_preview)
            if not ds.value(SETTINGS_KEYS["created_at"], ""):
                ds.setValue(SETTINGS_KEYS["created_at"], now_iso_utc())
            if not ds.value(SETTINGS_KEYS["is_quick_select"], None):
                ds.setValue(SETTINGS_KEYS["is_quick_select"], True)

        if write_user:
            us = open_user_settings(name, namespace=namespace)
            self._write_snapshot_to_settings(us, save_preview=save_preview)
            if not us.value(SETTINGS_KEYS["created_at"], ""):
                us.setValue(SETTINGS_KEYS["created_at"], now_iso_utc())
            if not us.value(SETTINGS_KEYS["is_quick_select"], None):
                us.setValue(SETTINGS_KEYS["is_quick_select"], True)

    def _finalize_profile_change(self, name: str, namespace: str | None) -> None:
        """
        Finalize a profile change by updating state and refreshing the UI.

        Args:
            name: The profile name.
            namespace: The profile namespace.
        """
        self._empty_profile_active = False
        self._empty_profile_consumed = True
        self._current_profile_name = name
        self.profile_changed.emit(name)
        set_last_profile(name, namespace=namespace, instance=self._last_profile_instance_id())
        combo = self.toolbar.components.get_action("workspace_combo").widget
        combo.refresh_profiles(active_profile=name)

    def _enter_empty_profile_state(self) -> None:
        """
        Switch to the transient empty workspace state.

        In this mode there is no active profile name, the toolbar shows an
        explicit blank profile entry, and no autosave on shutdown is performed.
        """
        self._empty_profile_active = True
        self._current_profile_name = None
        self._pending_autosave_skip = None
        self._refresh_workspace_list()

    @SafeSlot()
    def list_profiles(self) -> list[str]:
        """
        List available workspace profiles in the current namespace.

        Returns:
            list[str]: List of profile names.
        """
        namespace = self.profile_namespace
        return list_profiles(namespace)

    @SafeSlot(str)
    @rpc_timeout(None)
    def save_profile(
        self,
        name: str | None = None,
        *,
        show_dialog: bool = False,
        quick_select: bool | None = None,
    ):
        """
        Save the current workspace profile.

        On first save of a given name:
          - writes a default copy to states/default/<name>.ini with tag=default and created_at
          - writes a user copy   to states/user/<name>.ini    with tag=user    and created_at
        On subsequent saves of user-owned profiles:
          - updates both the default and user copies so restore uses the latest snapshot.
        Read-only bundled profiles cannot be overwritten.

        Args:
            name (str | None): The name of the profile to save. If None and show_dialog is True,
                prompts the user.
            show_dialog (bool): If True, shows the SaveProfileDialog for user interaction.
                If False (default), saves directly without user interaction (useful for CLI usage).
            quick_select (bool | None): Whether to include the profile in quick selection.
                If None (default), uses the existing value or True for new profiles.
                Only used when show_dialog is False; otherwise the dialog provides the value.
        """

        namespace = self.profile_namespace
        current_profile = getattr(self, "_current_profile_name", "") or ""

        def _profile_exists(profile_name: str) -> bool:
            return profile_origin(profile_name, namespace=namespace) != "unknown"

        # Determine final values either from dialog or directly
        if show_dialog:
            initial_name = name or ""
            quickselect_default = is_quick_select(name, namespace=namespace) if name else True

            dialog = SaveProfileDialog(
                self,
                current_name=initial_name,
                current_profile_name=current_profile,
                name_exists=_profile_exists,
                profile_origin=lambda n: profile_origin(n, namespace=namespace),
                origin_label=lambda n: profile_origin_display(n, namespace=namespace),
                quick_select_checked=quickselect_default,
            )
            if dialog.exec() != QDialog.DialogCode.Accepted:
                return

            name = dialog.get_profile_name()
            quickselect = dialog.is_quick_select()
            overwrite_existing = dialog.overwrite_existing
        else:
            # CLI / programmatic usage - no dialog
            if not name:
                logger.warning("save_profile called without name and show_dialog=False")
                return

            # Determine quick_select value
            if quick_select is None:
                # Use existing value if profile exists, otherwise default to True
                quickselect = (
                    is_quick_select(name, namespace=namespace) if _profile_exists(name) else True
                )
            else:
                quickselect = quick_select

            # For programmatic saves, check if profile is read-only
            origin = profile_origin(name, namespace=namespace)
            if origin in {"module", "plugin"}:
                logger.warning(f"Cannot save to read-only profile '{name}' (origin: {origin})")
                return

            # Overwrite existing settings profile when saving programmatically
            overwrite_existing = origin == "settings"

        origin_before_save = profile_origin(name, namespace=namespace)
        overwrite_default = overwrite_existing and origin_before_save == "settings"

        # Display saving placeholder in toolbar
        workspace_combo = self.toolbar.components.get_action("workspace_combo").widget
        workspace_combo.blockSignals(True)
        workspace_combo.insertItem(0, f"{name}-saving")
        workspace_combo.setCurrentIndex(0)
        workspace_combo.blockSignals(False)

        # Write to default and/or user settings
        should_write_default = overwrite_default or not any(
            os.path.exists(path) for path in default_profile_candidates(name, namespace)
        )
        self._write_profile_settings(
            name, namespace, write_default=should_write_default, write_user=True
        )

        set_quick_select(name, quickselect, namespace=namespace)

        self._refresh_workspace_list()
        if current_profile and current_profile != name and not overwrite_existing:
            self._pending_autosave_skip = (current_profile, name)
        else:
            self._pending_autosave_skip = None
        workspace_combo.setCurrentText(name)
        self._finalize_profile_change(name, namespace)

    @SafeSlot()
    @SafeSlot(str)
    def save_profile_dialog(self, name: str | None = None):
        """
        Save the current workspace profile with a dialog prompt.

        This is a convenience method for UI usage (toolbar, dialogs) that
        always shows the SaveProfileDialog. For programmatic/CLI usage,
        use save_profile() directly.

        Args:
            name (str | None): Optional initial name to populate in the dialog.
        """
        self.save_profile(name, show_dialog=True)

    @SafeSlot()
    @SafeSlot(str)
    @rpc_timeout(None)
    def load_profile(self, name: str | None = None):
        """
        Load a workspace profile.

        Before switching, persist the current profile to the user copy.
        Prefer loading the user copy; fall back to the default copy.

        Args:
            name (str | None): The name of the profile to load. If None, prompts the user.
        """
        if name == "":
            return

        if not name:  # Gui fallback if the name is not provided
            name, ok = QInputDialog.getText(
                self, "Load Workspace", "Enter the name of the workspace profile to load:"
            )
            if not ok or not name:
                return

        namespace = self.profile_namespace
        prev_name = getattr(self, "_current_profile_name", None)
        skip_pair = getattr(self, "_pending_autosave_skip", None)
        if prev_name and prev_name != name:
            if skip_pair and skip_pair == (prev_name, name):
                self._pending_autosave_skip = None
            else:
                us_prev = open_user_settings(prev_name, namespace=namespace)
                self._write_snapshot_to_settings(us_prev, save_preview=True)

        settings = None
        if any(os.path.exists(path) for path in user_profile_candidates(name, namespace)):
            settings = open_user_settings(name, namespace=namespace)
        elif any(os.path.exists(path) for path in default_profile_candidates(name, namespace)):
            settings = open_default_settings(name, namespace=namespace)
        if settings is None:
            logger.warning(f"Profile '{name}' not found in namespace '{namespace}'. Creating new.")
            self.delete_all()
            self.save_profile(name, show_dialog=False, quick_select=True)
            return

        # Clear existing docks and remove all widgets
        self.delete_all()

        # Rebuild widgets and restore states
        for item in read_manifest(settings):
            obj_name = item["object_name"]
            widget_class = item["widget_class"]
            if obj_name not in self.widget_map():
                w = widget_handler.create_widget(widget_type=widget_class, parent=self)
                w.setObjectName(obj_name)
                floating_state = None
                if item.get("floating"):
                    floating_state = {
                        "relative": item.get("floating_relative"),
                        "absolute": item.get("floating_absolute"),
                        "screen_name": item.get("floating_screen"),
                    }
                self._make_dock(
                    w,
                    closable=item["closable"],
                    floatable=item["floatable"],
                    movable=item["movable"],
                    start_floating=item.get("floating", False),
                    floating_state=floating_state,
                    area=QtAds.DockWidgetArea.RightDockWidgetArea,
                )

        self.load_from_settings(settings, keys=PROFILE_STATE_KEYS)
        self.state_manager.load_state(settings=settings)
        self._set_editable(self._editable)

        self._finalize_profile_change(name, namespace)

    @SafeSlot()
    @SafeSlot(str)
    def restore_user_profile_from_default(self, name: str | None = None):
        """
        Overwrite the user copy of *name* with the default baseline.
        If *name* is None, target the currently active profile.

        Args:
            name (str | None): The name of the profile to restore. If None, uses the current profile.
        """
        target = name or getattr(self, "_current_profile_name", None)
        if not target:
            return
        namespace = self.profile_namespace

        current_pixmap = None
        if self.isVisible():
            current_pixmap = QPixmap()
            ba = bytes(self.screenshot_bytes())
            current_pixmap.loadFromData(ba)
        if current_pixmap is None or current_pixmap.isNull():
            current_pixmap = load_user_profile_screenshot(target, namespace=namespace)
        default_pixmap = load_default_profile_screenshot(target, namespace=namespace)

        if not RestoreProfileDialog.confirm(self, current_pixmap, default_pixmap):
            return

        restore_user_from_default(target, namespace=namespace)
        self.delete_all()
        self.load_profile(target)

    @SafeSlot()
    def delete_profile(self, name: str | None = None, show_dialog: bool = False) -> bool:
        """
        Delete a workspace profile.

        Args:
            name: The name of the profile to delete. If None, uses the currently
                selected profile from the toolbar combo box (for UI usage).
            show_dialog: If True, show confirmation dialog before deletion.
                Defaults to False for CLI/programmatic usage.

        Returns:
            bool: True if the profile was deleted, False otherwise.

        Raises:
            ValueError: If the profile is read-only or doesn't exist (when show_dialog=False).

        """
        # Resolve profile name
        if name is None:
            combo = self.toolbar.components.get_action("workspace_combo").widget
            name = combo.currentText()
        if not name:
            if show_dialog:
                return False
            raise ValueError("No profile name provided.")

        namespace = self.profile_namespace

        # Check if profile is read-only
        if is_profile_read_only(name, namespace=namespace):
            if show_dialog:
                QMessageBox.information(
                    self, "Delete Profile", f"Profile '{name}' is read-only and cannot be deleted."
                )
                return False
            raise ValueError(f"Profile '{name}' is read-only and cannot be deleted.")

        # Confirm deletion if dialog is enabled
        if show_dialog:
            reply = QMessageBox.question(
                self,
                "Delete Profile",
                f"Are you sure you want to delete the profile '{name}'?\n\n"
                f"This action cannot be undone.",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            if reply != QMessageBox.StandardButton.Yes:
                return False

        # Perform deletion
        try:
            removed = delete_profile_files(name, namespace=namespace)
        except OSError as exc:
            if show_dialog:
                QMessageBox.warning(
                    self, "Delete Profile", f"Failed to delete profile '{name}': {exc}"
                )
                return False
            raise ValueError(f"Failed to delete profile '{name}': {exc}") from exc

        if not removed:
            if show_dialog:
                QMessageBox.information(
                    self, "Delete Profile", "No writable profile files were found to delete."
                )
                return False
            raise ValueError(f"No writable profile files found for '{name}'.")

        # Clear current profile if it was deleted
        if getattr(self, "_current_profile_name", None) == name:
            self._current_profile_name = None

        # Refresh the workspace list
        self._refresh_workspace_list()
        return True

    def _refresh_workspace_list(self):
        """
        Populate the workspace combo box with all saved profile names (without .ini).
        """
        combo = self.toolbar.components.get_action("workspace_combo").widget
        active_profile = getattr(self, "_current_profile_name", None)
        empty_profile_active = bool(getattr(self, "_empty_profile_active", False))
        namespace = self.profile_namespace
        if hasattr(combo, "set_quick_profile_provider"):
            combo.set_quick_profile_provider(lambda ns=namespace: list_quick_profiles(namespace=ns))
        if hasattr(combo, "refresh_profiles"):
            if empty_profile_active:
                combo.refresh_profiles(active_profile, show_empty_profile=True)
            else:
                combo.refresh_profiles(active_profile)
        else:
            # Fallback for regular QComboBox
            combo.blockSignals(True)
            combo.clear()
            quick_profiles = list_quick_profiles(namespace=namespace)
            items = [""] if empty_profile_active else []
            items.extend(quick_profiles)
            if active_profile and active_profile not in items:
                items.insert(0, active_profile)
            combo.addItems(items)
            if empty_profile_active:
                idx = combo.findText("")
                if idx >= 0:
                    combo.setCurrentIndex(idx)
            elif active_profile:
                idx = combo.findText(active_profile)
                if idx >= 0:
                    combo.setCurrentIndex(idx)
            if empty_profile_active:
                combo.setToolTip("Unsaved empty workspace")
            elif active_profile and active_profile not in quick_profiles:
                combo.setToolTip("Active profile is not in quick select")
            else:
                combo.setToolTip("")
            combo.blockSignals(False)

    ################################################################################
    # Dialog Popups
    ################################################################################

    @SafeSlot()
    def show_workspace_manager(self):
        """
        Show the workspace manager dialog.
        """
        manage_action = self.toolbar.components.get_action("manage_workspaces").action
        if self.manage_dialog is None or not self.manage_dialog.isVisible():
            self.manage_widget = WorkSpaceManager(
                self, target_widget=self, default_profile=self._current_profile_name
            )
            self.manage_dialog = QDialog(modal=False)

            self.manage_dialog.setWindowTitle("Workspace Manager")
            self.manage_dialog.setMinimumSize(1200, 500)
            self.manage_dialog.layout = QVBoxLayout(self.manage_dialog)
            self.manage_dialog.layout.addWidget(self.manage_widget)
            self.manage_dialog.finished.connect(self._manage_dialog_closed)
            self.manage_dialog.show()
            self.manage_dialog.resize(300, 300)
            manage_action.setChecked(True)
        else:
            # If already open, bring it to the front
            self.manage_dialog.raise_()
            self.manage_dialog.activateWindow()
            manage_action.setChecked(True)  # keep it toggle

    def _manage_dialog_closed(self):
        self.manage_widget.close()
        self.manage_widget.deleteLater()
        if self.manage_dialog is not None:
            self.manage_dialog.deleteLater()
            self.manage_dialog = None
        self.toolbar.components.get_action("manage_workspaces").action.setChecked(False)

    ################################################################################
    # Mode Switching
    ################################################################################

    @SafeProperty(str)
    def mode(self) -> str:
        return self._mode

    @mode.setter
    def mode(self, new_mode: str):
        allowed_modes = ["plot", "device", "utils", "user", "creator"]
        if new_mode not in allowed_modes:
            raise ValueError(f"Invalid mode: {new_mode}")
        self._mode = new_mode
        self.mode_changed.emit(new_mode)
        self._apply_toolbar_layout()

    def _apply_toolbar_layout(self) -> None:
        mode_key = getattr(self, "_mode", "creator")
        if mode_key == "user":
            bundles = ["spacer_bundle", "workspace", "dock_actions"]
        elif mode_key == "creator":
            bundles = [
                "menu_plots",
                "menu_devices",
                "menu_utils",
                "spacer_bundle",
                "workspace",
                "dock_actions",
            ]
        elif mode_key == "plot":
            bundles = ["flat_plots", "spacer_bundle", "workspace", "dock_actions"]
        elif mode_key == "device":
            bundles = ["flat_devices", "spacer_bundle", "workspace", "dock_actions"]
        elif mode_key == "utils":
            bundles = ["flat_utils", "spacer_bundle", "workspace", "dock_actions"]
        else:
            bundles = ["spacer_bundle", "workspace", "dock_actions"]

        if not self._profile_management_enabled:
            flat_only = [b for b in bundles if b.startswith("flat_")]
            if not flat_only:
                flat_only = ["flat_plots", "flat_devices", "flat_utils"]
            bundles = flat_only

        self.toolbar.show_bundles(bundles)

    def prepare_for_shutdown(self) -> None:
        """
        Persist the current workspace snapshot while the UI is still fully visible.
        Called by the main window before initiating widget teardown to avoid capturing
        close-triggered visibility changes.
        """
        if (
            not self._auto_save_upon_exit
            or getattr(self, "_exit_snapshot_written", False)
            or getattr(self, "_destroyed", False)
        ):
            logger.info("ADS prepare_for_shutdown: skipping (already handled or destroyed)")
            return

        if getattr(self, "_empty_profile_active", False):
            logger.info("ADS prepare_for_shutdown: skipping autosave for unsaved empty workspace")
            self._exit_snapshot_written = True
            return

        name = getattr(self, "_current_profile_name", None)
        if not name:
            logger.info("ADS prepare_for_shutdown: skipping autosave (no active profile)")
            self._exit_snapshot_written = True
            return

        namespace = self.profile_namespace
        settings = open_user_settings(name, namespace=namespace)
        self._write_snapshot_to_settings(settings)
        set_last_profile(name, namespace=namespace, instance=self._last_profile_instance_id())
        self._exit_snapshot_written = True

    def register_tour_steps(self, guided_tour, main_app):
        """Register Dock Area components with the guided tour.

        Args:
            guided_tour: The GuidedTour instance to register with.
            main_app: The main application instance (for accessing set_current).

        Returns:
            ViewTourSteps | None: Model containing view title and step IDs.
        """

        step_ids = []

        # Register Dock Area toolbar
        def get_dock_toolbar():
            main_app.set_current("dock_area")
            return (self.toolbar, None)

        step_id = guided_tour.register_widget(
            widget=get_dock_toolbar,
            title="Dock Area Toolbar",
            text="Use this toolbar to add widgets, manage workspaces, save and load profiles, and control the layout of your workspace.",
        )
        step_ids.append(step_id)

        return ViewTourSteps(view_title="Dock Area Workspace", step_ids=step_ids)

    def cleanup(self):
        """
        Cleanup the dock area.
        """
        self.prepare_for_shutdown()
        if self.manage_dialog is not None:
            self.manage_dialog.reject()
            self.manage_dialog = None
        self.delete_all()
        self.dark_mode_button.close()
        self.dark_mode_button.deleteLater()
        self.toolbar.cleanup()
        super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    import sys

    app = QApplication(sys.argv)
    apply_theme("dark")
    dispatcher = BECDispatcher(gui_id="ads")
    window = BECMainWindowNoRPC()

    ads = BECDockArea(mode="creator", enable_profile_management=True, root_widget=True)

    window.setCentralWidget(ads)
    window.show()
    window.resize(800, 1000)

    sys.exit(app.exec())



from __future__ import annotations

import collections
import enum
import typing

from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Signal

from bec_widgets.widgets.containers.qt_ads import ads
from bec_widgets.widgets.containers.qt_ads.ads import *

# pylint: disable=unused-argument,invalid-name, missing-function-docstring, super-init-not-called

class CAutoHideDockContainer(QtWidgets.QFrame):
    def __init__(
        self,
        DockWidget: typing.Optional["CDockWidget"],
        area: SideBarLocation,
        parent: typing.Optional["CDockContainerWidget"],
    ) -> None: ...
    def moveToNewSideBarLocation(self, a0: SideBarLocation) -> None: ...
    def orientation(self) -> QtCore.Qt.Orientation: ...
    def resetToInitialDockWidgetSize(self) -> None: ...
    def setSize(self, Size: int) -> None: ...
    def toggleCollapseState(self) -> None: ...
    def collapseView(self, Enable: bool) -> None: ...
    def toggleView(self, Enable: bool) -> None: ...
    def cleanupAndDelete(self) -> None: ...
    def moveContentsToParent(self) -> None: ...
    def dockContainer(self) -> typing.Optional["CDockContainerWidget"]: ...
    def dockAreaWidget(self) -> typing.Optional["CDockAreaWidget"]: ...
    def setSideBarLocation(self, SideBarLocation: SideBarLocation) -> None: ...
    def sideBarLocation(self) -> SideBarLocation: ...
    def addDockWidget(self, DockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def tabIndex(self) -> int: ...
    def dockWidget(self) -> typing.Optional["CDockWidget"]: ...
    def autoHideTab(self) -> typing.Optional["CAutoHideTab"]: ...
    def autoHideSideBar(self) -> typing.Optional["CAutoHideSideBar"]: ...
    def saveState(self, Stream: QtCore.QXmlStreamWriter) -> None: ...
    def updateSize(self) -> None: ...
    def event(self, event: typing.Optional[QtCore.QEvent]) -> bool: ...
    def leaveEvent(self, event: typing.Optional[QtCore.QEvent]) -> None: ...
    def resizeEvent(self, event: typing.Optional[QtGui.QResizeEvent]) -> None: ...
    def eventFilter(
        self, watched: typing.Optional[QtCore.QObject], event: typing.Optional[QtCore.QEvent]
    ) -> bool: ...

class CAutoHideSideBar(QtWidgets.QScrollArea):
    def __init__(
        self, parent: typing.Optional["CDockContainerWidget"], area: "SideBarLocation"
    ) -> None: ...
    def dockContainer(self) -> typing.Optional["CDockContainerWidget"]: ...
    def setSpacing(self, Spacing: int) -> None: ...
    def spacing(self) -> int: ...
    def sizeHint(self) -> QtCore.QSize: ...
    def minimumSizeHint(self) -> QtCore.QSize: ...
    def sideBarLocation(self) -> "SideBarLocation": ...
    def hasVisibleTabs(self) -> bool: ...
    def visibleTabCount(self) -> int: ...
    def count(self) -> int: ...
    def indexOfTab(self, Tab: "CAutoHideTab") -> int: ...
    def tabInsertIndexAt(self, Pos: QtCore.QPoint) -> int: ...
    def tabAt(self, Pos: QtCore.QPoint) -> int: ...
    def tab(self, index: int) -> typing.Optional["CAutoHideTab"]: ...
    def orientation(self) -> QtCore.Qt.Orientation: ...
    def addAutoHideWidget(
        self, AutoHideWidget: typing.Optional["CAutoHideDockContainer"], Index: int
    ) -> None: ...
    def removeAutoHideWidget(
        self, AutoHideWidget: typing.Optional["CAutoHideDockContainer"]
    ) -> None: ...
    def insertDockWidget(
        self, Index: int, DockWidget: typing.Optional["CDockWidget"]
    ) -> typing.Optional["CAutoHideDockContainer"]: ...
    def removeTab(self, SideTab: typing.Optional["CAutoHideTab"]) -> None: ...
    def insertTab(self, Index: int, SideTab: typing.Optional["CAutoHideTab"]) -> None: ...
    def saveState(self, Stream: QtCore.QXmlStreamWriter) -> None: ...
    def eventFilter(
        self, watched: typing.Optional[QtCore.QObject], event: typing.Optional[QtCore.QEvent]
    ) -> bool: ...

class CPushButton(QtWidgets.QPushButton):
    class Orientation(enum.Enum):
        Horizontal = ...
        VerticalTopToBottom = ...
        VerticalBottomToTop = ...

    def __init__(self) -> None: ...
    def setButtonOrientation(self, orientation: "CPushButton.Orientation") -> None: ...
    def buttonOrientation(self) -> "CPushButton.Orientation": ...
    def sizeHint(self) -> QtCore.QSize: ...

class CAutoHideTab(CPushButton):
    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = ...) -> None: ...
    def requestCloseDockWidget(self) -> None: ...
    def unpinDockWidget(self) -> None: ...
    def setDockWidgetFloating(self) -> None: ...
    def tabIndex(self) -> int: ...
    def sideBar(self) -> typing.Optional["CAutoHideSideBar"]: ...
    def iconOnly(self) -> bool: ...
    def setDockWidget(self, DockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def dockWidget(self) -> typing.Optional["CDockWidget"]: ...
    def isActiveTab(self) -> bool: ...
    def orientation(self) -> QtCore.Qt.Orientation: ...
    def setOrientation(self, Orientation: QtCore.Qt.Orientation) -> None: ...
    def sideBarLocation(self) -> "SideBarLocation": ...
    def updateStyle(self) -> None: ...
    def dragLeaveEvent(self, ev: typing.Optional[QtGui.QDragLeaveEvent]) -> None: ...
    def dragEnterEvent(self, ev: typing.Optional[QtGui.QDragEnterEvent]) -> None: ...
    def mouseMoveEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mouseReleaseEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mousePressEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def contextMenuEvent(self, ev: typing.Optional[QtGui.QContextMenuEvent]) -> None: ...
    def event(self, event: typing.Optional[QtCore.QEvent]) -> bool: ...
    def removeFromSideBar(self) -> None: ...
    def setSideBar(self, SideTabBar: typing.Optional["CAutoHideSideBar"]) -> None: ...

class CDockWidget(QtWidgets.QFrame):
    class eToggleViewActionMode(enum.Enum):
        ActionModeToggle = ...
        ActionModeShow = ...

    class eMinimumSizeHintMode(enum.Enum):
        MinimumSizeHintFromDockWidget = ...
        MinimumSizeHintFromContent = ...
        MinimumSizeHintFromDockWidgetMinimumSize = ...
        MinimumSizeHintFromContentMinimumSize = ...

    class eInsertMode(enum.Enum):
        AutoScrollArea = ...
        ForceScrollArea = ...
        ForceNoScrollArea = ...

    class eToolBarStyleSource(enum.Enum):
        ToolBarStyleFromDockManager = ...
        ToolBarStyleFromDockWidget = ...

    class eState(enum.Enum):
        StateHidden = ...
        StateDocked = ...
        StateFloating = ...

    class DockWidgetFeature(enum.Enum):
        DockWidgetClosable = ...
        DockWidgetMovable = ...
        DockWidgetFloatable = ...
        DockWidgetDeleteOnClose = ...
        CustomCloseHandling = ...
        DockWidgetFocusable = ...
        DockWidgetForceCloseWithArea = ...
        NoTab = ...
        DeleteContentOnClose = ...
        DockWidgetPinnable = ...
        DefaultDockWidgetFeatures = ...
        AllDockWidgetFeatures = ...
        DockWidgetAlwaysCloseAndDelete = ...
        GloballyLockableFeatures = ...
        NoDockWidgetFeatures = ...

    @typing.overload
    def __init__(
        self, title: typing.Optional[str], parent: typing.Optional[QtWidgets.QWidget] = ...
    ) -> None: ...
    @typing.overload
    def __init__(
        self,
        manager: typing.Optional["CDockManager"],
        title: typing.Optional[str],
        parent: typing.Optional[QtWidgets.QWidget] = ...,
    ) -> None: ...

    featuresChanged: typing.ClassVar[Signal]
    visibilityChanged: typing.ClassVar[Signal]
    closeRequested: typing.ClassVar[Signal]
    topLevelChanged: typing.ClassVar[Signal]
    titleChanged: typing.ClassVar[Signal]
    closed: typing.ClassVar[Signal]
    viewToggled: typing.ClassVar[Signal]
    def toggleAutoHide(self, Location: "SideBarLocation" = ...) -> None: ...
    def setAutoHide(
        self, Enable: bool, Location: "SideBarLocation" = ..., TabIndex: int = ...
    ) -> None: ...
    def showNormal(self) -> None: ...
    def showFullScreen(self) -> None: ...
    def requestCloseDockWidget(self) -> None: ...
    def closeDockWidget(self) -> None: ...
    def deleteDockWidget(self) -> None: ...
    def setFloating(self) -> None: ...
    def raise_(self) -> None: ...
    def setAsCurrentTab(self) -> None: ...
    def toggleView(self, Open: bool = ...) -> None: ...
    def event(self, e: typing.Optional[QtCore.QEvent]) -> bool: ...
    def isCurrentTab(self) -> bool: ...
    def isTabbed(self) -> bool: ...
    def isFullScreen(self) -> bool: ...
    def setTabToolTip(self, text: typing.Optional[str]) -> None: ...
    def titleBarActions(self) -> list[QtGui.QAction]: ...
    def setTitleBarActions(self, actions: collections.abc.Iterable[QtGui.QAction]) -> None: ...
    def toolBarIconSize(self, State: "CDockWidget.eState") -> QtCore.QSize: ...
    def setToolBarIconSize(self, IconSize: QtCore.QSize, State: "CDockWidget.eState") -> None: ...
    def toolBarStyle(self, State: "CDockWidget.eState") -> QtCore.Qt.ToolButtonStyle: ...
    def setToolBarStyle(
        self, Style: QtCore.Qt.ToolButtonStyle, State: "CDockWidget.eState"
    ) -> None: ...
    def toolBarStyleSource(self) -> "CDockWidget.eToolBarStyleSource": ...
    def setToolBarStyleSource(self, Source: "CDockWidget.eToolBarStyleSource") -> None: ...
    def setToolBar(self, ToolBar: typing.Optional[QtWidgets.QToolBar]) -> None: ...
    def createDefaultToolBar(self) -> typing.Optional[QtWidgets.QToolBar]: ...
    def toolBar(self) -> typing.Optional[QtWidgets.QToolBar]: ...
    def icon(self) -> QtGui.QIcon: ...
    def setIcon(self, Icon: QtGui.QIcon) -> None: ...
    def isCentralWidget(self) -> bool: ...
    def minimumSizeHintMode(self) -> "CDockWidget.eMinimumSizeHintMode": ...
    def setMinimumSizeHintMode(self, Mode: "CDockWidget.eMinimumSizeHintMode") -> None: ...
    def setToggleViewActionMode(self, Mode: "CDockWidget.eToggleViewActionMode") -> None: ...
    def setToggleViewAction(self, action: typing.Optional[QtGui.QAction]) -> None: ...
    def toggleViewAction(self) -> typing.Optional[QtGui.QAction]: ...
    def isClosed(self) -> bool: ...
    def isInFloatingContainer(self) -> bool: ...
    def isFloating(self) -> bool: ...
    def autoHideLocation(self) -> "SideBarLocation": ...
    def autoHideDockContainer(self) -> typing.Optional["CAutoHideDockContainer"]: ...
    def isAutoHide(self) -> bool: ...
    def setSideTabWidget(self, SideTab: typing.Optional["CAutoHideTab"]) -> None: ...
    def sideTabWidget(self) -> typing.Optional["CAutoHideTab"]: ...
    def dockAreaWidget(self) -> typing.Optional["CDockAreaWidget"]: ...
    def floatingDockContainer(self) -> typing.Optional["CFloatingDockContainer"]: ...
    def dockContainer(self) -> typing.Optional["CDockContainerWidget"]: ...
    def dockManager(self) -> typing.Optional["CDockManager"]: ...
    def features(self) -> "CDockWidget.DockWidgetFeature": ...
    def setFeature(self, flag: "CDockWidget.DockWidgetFeature", on: bool) -> None: ...
    def setFeatures(self, features: "CDockWidget.DockWidgetFeature") -> None: ...
    def tabWidget(self) -> typing.Optional["CDockWidgetTab"]: ...
    def widget(self) -> typing.Optional[QtWidgets.QWidget]: ...
    def takeWidget(self) -> typing.Optional[QtWidgets.QWidget]: ...
    def setWidget(
        self,
        widget: typing.Optional[QtWidgets.QWidget],
        InsertMode: "CDockWidget.eInsertMode" = ...,
    ) -> None: ...
    def minimumSizeHint(self) -> QtCore.QSize: ...
    def closeDockWidgetInternal(self, ForceClose: bool = ...) -> bool: ...
    def toggleViewInternal(self, Open: bool) -> None: ...
    def setClosedState(self, Closed: bool) -> None: ...
    def emitTopLevelChanged(self, Floating: bool) -> None: ...
    @staticmethod
    def emitTopLevelEventForWidget(
        TopLevelDockWidget: typing.Optional["CDockWidget"], Floating: bool
    ) -> None: ...
    def flagAsUnassigned(self) -> None: ...
    def saveState(self, Stream: QtCore.QXmlStreamWriter) -> None: ...
    def setToggleViewActionChecked(self, Checked: bool) -> None: ...
    def setDockArea(self, DockArea: typing.Optional["CDockAreaWidget"]) -> None: ...
    def setDockManager(self, DockManager: typing.Optional["CDockManager"]) -> None: ...

class CDockAreaTabBar(QtWidgets.QScrollArea):
    def __init__(self, parent: typing.Optional["CDockAreaWidget"]) -> None: ...

    elidedChanged: typing.ClassVar[Signal]
    tabInserted: typing.ClassVar[Signal]
    removingTab: typing.ClassVar[Signal]
    tabMoved: typing.ClassVar[Signal]
    tabOpened: typing.ClassVar[Signal]
    tabClosed: typing.ClassVar[Signal]
    tabCloseRequested: typing.ClassVar[Signal]
    tabBarClicked: typing.ClassVar[Signal]
    currentChanged: typing.ClassVar[Signal]
    currentChanging: typing.ClassVar[Signal]
    def closeTab(self, Index: int) -> None: ...
    def setCurrentIndex(self, Index: int) -> None: ...
    def areTabsOverflowing(self) -> bool: ...
    def sizeHint(self) -> QtCore.QSize: ...
    def minimumSizeHint(self) -> QtCore.QSize: ...
    def isTabOpen(self, Index: int) -> bool: ...
    def eventFilter(
        self, watched: typing.Optional[QtCore.QObject], event: typing.Optional[QtCore.QEvent]
    ) -> bool: ...
    def tabInsertIndexAt(self, Pos: QtCore.QPoint) -> int: ...
    def tabAt(self, Pos: QtCore.QPoint) -> int: ...
    def tab(self, Index: int) -> typing.Optional["CDockWidgetTab"]: ...
    def currentTab(self) -> typing.Optional["CDockWidgetTab"]: ...
    def currentIndex(self) -> int: ...
    def count(self) -> int: ...
    def removeTab(self, Tab: typing.Optional["CDockWidgetTab"]) -> None: ...
    def insertTab(self, Index: int, Tab: typing.Optional["CDockWidgetTab"]) -> None: ...
    def wheelEvent(self, Event: typing.Optional[QtGui.QWheelEvent]) -> None: ...

class CSpacerWidget(QtWidgets.QWidget):
    def __init__(self, Parent: typing.Optional[QtWidgets.QWidget] = ...) -> None: ...
    def minimumSizeHint(self) -> QtCore.QSize: ...
    def sizeHint(self) -> QtCore.QSize: ...

class CTitleBarButton(QtWidgets.QToolButton):
    def __init__(
        self,
        ShowInTitleBar: bool,
        HideWhenDisabled: bool,
        ButtonId: "TitleBarButton",
        parent: typing.Optional[QtWidgets.QWidget] = ...,
    ) -> None: ...
    def event(self, ev: typing.Optional[QtCore.QEvent]) -> bool: ...
    def isInAutoHideArea(self) -> bool: ...
    def titleBar(self) -> typing.Optional["CDockAreaTitleBar"]: ...
    def buttonId(self) -> "TitleBarButton": ...
    def setShowInTitleBar(self, a0: bool) -> None: ...
    def setVisible(self, a0: bool) -> None: ...

class CDockAreaTitleBar(QtWidgets.QFrame):
    def __init__(self, parent: typing.Optional[CDockAreaWidget]) -> None: ...

    tabBarClicked: typing.ClassVar[Signal]
    def buildContextMenu(
        self, menu: typing.Optional[QtWidgets.QMenu] = ...
    ) -> typing.Optional[QtWidgets.QMenu]: ...
    def isAutoHide(self) -> bool: ...
    def showAutoHideControls(self, Show: bool) -> None: ...
    def setAreaFloating(self) -> None: ...
    def titleBarButtonToolTip(self, Button: "TitleBarButton") -> str: ...
    def indexOf(self, widget: typing.Optional[QtWidgets.QWidget]) -> int: ...
    def insertWidget(self, index: int, widget: typing.Optional[QtWidgets.QWidget]) -> None: ...
    def setVisible(self, Visible: bool) -> None: ...
    def updateDockWidgetActionsButtons(self) -> None: ...
    def dockAreaWidget(self) -> typing.Optional["CDockAreaWidget"]: ...
    def autoHideTitleLabel(self) -> typing.Optional["CElidingLabel"]: ...
    def button(self, which: "TitleBarButton") -> typing.Optional["CTitleBarButton"]: ...
    def tabBar(self) -> typing.Optional["CDockAreaTabBar"]: ...
    def markTabsMenuOutdated(self) -> None: ...
    def resizeEvent(self, event: typing.Optional[QtGui.QResizeEvent]) -> None: ...
    def contextMenuEvent(self, event: typing.Optional[QtGui.QContextMenuEvent]) -> None: ...
    def mouseDoubleClickEvent(self, event: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mouseMoveEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mouseReleaseEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mousePressEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...

class CDockAreaWidget(QtWidgets.QFrame):
    class eDockAreaFlag(enum.Enum):
        HideSingleWidgetTitleBar = ...
        DefaultFlags = ...

    def __init__(
        self,
        DockManager: typing.Optional[CDockManager],
        parent: typing.Optional[CDockContainerWidget],
    ) -> None: ...

    viewToggled: typing.ClassVar[Signal]
    currentChanged: typing.ClassVar[Signal]
    currentChanging: typing.ClassVar[Signal]
    tabBarClicked: typing.ClassVar[Signal]
    def setFloating(self) -> None: ...
    def closeOtherAreas(self) -> None: ...
    def toggleAutoHide(self, Location: "SideBarLocation" = ...) -> None: ...
    def setAutoHide(
        self, Enable: bool, Location: "SideBarLocation" = ..., TabIndex: int = ...
    ) -> None: ...
    def closeArea(self) -> None: ...
    def setCurrentIndex(self, index: int) -> None: ...
    def isTopLevelArea(self) -> bool: ...
    def containsCentralWidget(self) -> bool: ...
    def isCentralWidgetArea(self) -> bool: ...
    def setDockAreaFlag(self, Flag: "CDockAreaWidget.eDockAreaFlag", On: bool) -> None: ...
    def setDockAreaFlags(self, Flags: "CDockAreaWidget.eDockAreaFlag") -> None: ...
    def dockAreaFlags(self) -> "CDockAreaWidget.eDockAreaFlag": ...
    def titleBar(self) -> typing.Optional["CDockAreaTitleBar"]: ...
    def allowedAreas(self) -> "DockWidgetArea": ...
    def setAllowedAreas(self, areas: "DockWidgetArea") -> None: ...
    def setVisible(self, Visible: bool) -> None: ...
    def titleBarButton(
        self, which: "TitleBarButton"
    ) -> typing.Optional[QtWidgets.QAbstractButton]: ...
    def features(self, Mode: "eBitwiseOperator" = ...) -> "CDockWidget.DockWidgetFeature": ...
    @staticmethod
    def restoreState(
        Stream: "CDockingStateReader",
        Testing: bool,
        ParentContainer: typing.Optional["CDockContainerWidget"],
    ) -> typing.Tuple[bool, typing.Optional["CDockAreaWidget"]]: ...
    def saveState(self, Stream: QtCore.QXmlStreamWriter) -> None: ...
    def setCurrentDockWidget(self, DockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def currentDockWidget(self) -> typing.Optional["CDockWidget"]: ...
    def indexOfFirstOpenDockWidget(self) -> int: ...
    def currentIndex(self) -> int: ...
    def dockWidget(self, Index: int) -> typing.Optional["CDockWidget"]: ...
    def openedDockWidgets(self) -> list["CDockWidget"]: ...
    def openDockWidgetsCount(self) -> int: ...
    def dockWidgets(self) -> list["CDockWidget"]: ...
    def dockWidgetsCount(self) -> int: ...
    def contentAreaGeometry(self) -> QtCore.QRect: ...
    def titleBarGeometry(self) -> QtCore.QRect: ...
    def minimumSizeHint(self) -> QtCore.QSize: ...
    def setAutoHideDockContainer(self, a0: typing.Optional["CAutoHideDockContainer"]) -> None: ...
    def isAutoHide(self) -> bool: ...
    def parentSplitter(self) -> typing.Optional["CDockSplitter"]: ...
    def autoHideDockContainer(self) -> typing.Optional["CAutoHideDockContainer"]: ...
    def dockContainer(self) -> typing.Optional["CDockContainerWidget"]: ...
    def dockManager(self) -> typing.Optional["CDockManager"]: ...
    def toggleView(self, Open: bool) -> None: ...
    def updateTitleBarButtonVisibility(self, IsTopLevel: bool) -> None: ...
    def markTitleBarMenuOutdated(self) -> None: ...
    def internalSetCurrentDockWidget(self, DockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def updateTitleBarVisibility(self) -> None: ...
    def hideAreaWithNoVisibleContent(self) -> None: ...
    def index(self, DockWidget: typing.Optional["CDockWidget"]) -> int: ...
    def nextOpenDockWidget(
        self, DockWidget: typing.Optional["CDockWidget"]
    ) -> typing.Optional["CDockWidget"]: ...
    def toggleDockWidgetView(
        self, DockWidget: typing.Optional["CDockWidget"], Open: bool
    ) -> None: ...
    def removeDockWidget(self, DockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def addDockWidget(self, DockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def insertDockWidget(
        self, index: int, DockWidget: typing.Optional["CDockWidget"], Activate: bool = ...
    ) -> None: ...

class CDockContainerWidget(QtWidgets.QFrame):
    def __init__(
        self,
        DockManager: typing.Optional["CDockManager"],
        parent: typing.Optional[QtWidgets.QWidget] = ...,
    ) -> None: ...

    dockAreaViewToggled: typing.ClassVar[Signal]
    dockAreasRemoved: typing.ClassVar[Signal]
    autoHideWidgetCreated: typing.ClassVar[Signal]
    dockAreasAdded: typing.ClassVar[Signal]
    def dockManager(self) -> typing.Optional["CDockManager"]: ...
    def contentRectGlobal(self) -> QtCore.QRect: ...
    def contentRect(self) -> QtCore.QRect: ...
    def autoHideWidgets(self) -> list["CAutoHideDockContainer"]: ...
    def autoHideSideBar(self, area: "SideBarLocation") -> typing.Optional["CAutoHideSideBar"]: ...
    def closeOtherAreas(self, KeepOpenArea: typing.Optional["CDockAreaWidget"]) -> None: ...
    def floatingWidget(self) -> typing.Optional["CFloatingDockContainer"]: ...
    def features(self) -> "CDockWidget.DockWidgetFeature": ...
    def dumpLayout(self) -> None: ...
    def isFloating(self) -> bool: ...
    def visibleDockAreaCount(self) -> int: ...
    def dockAreaCount(self) -> int: ...
    def hasTopLevelDockWidget(self) -> bool: ...
    def openedDockWidgets(self) -> list["CDockWidget"]: ...
    def openedDockAreas(self) -> list["CDockAreaWidget"]: ...
    def dockArea(self, Index: int) -> typing.Optional["CDockAreaWidget"]: ...
    def dockAreaAt(self, GlobalPos: QtCore.QPoint) -> typing.Optional["CDockAreaWidget"]: ...
    def isInFrontOf(self, Other: typing.Optional["CDockContainerWidget"]) -> bool: ...
    def zOrderIndex(self) -> int: ...
    def removeDockWidget(self, Dockwidget: typing.Optional["CDockWidget"]) -> None: ...
    def addDockWidget(
        self,
        area: "DockWidgetArea",
        Dockwidget: typing.Optional["CDockWidget"],
        DockAreaWidget: typing.Optional["CDockAreaWidget"] = ...,
        Index: int = ...,
    ) -> typing.Optional["CDockAreaWidget"]: ...
    def handleAutoHideWidgetEvent(
        self, e: typing.Optional[QtCore.QEvent], w: typing.Optional[QtWidgets.QWidget]
    ) -> None: ...
    def removeAutoHideWidget(
        self, AutoHideWidget: typing.Optional["CAutoHideDockContainer"]
    ) -> None: ...
    def registerAutoHideWidget(
        self, AutoHideWidget: typing.Optional["CAutoHideDockContainer"]
    ) -> None: ...
    def updateSplitterHandles(self, splitter: typing.Optional[QtWidgets.QSplitter]) -> None: ...
    def dockWidgets(self) -> list["CDockWidget"]: ...
    def topLevelDockArea(self) -> typing.Optional["CDockAreaWidget"]: ...
    def topLevelDockWidget(self) -> typing.Optional["CDockWidget"]: ...
    def lastAddedDockAreaWidget(
        self, area: "DockWidgetArea"
    ) -> typing.Optional["CDockAreaWidget"]: ...
    def restoreState(self, Stream: "CDockingStateReader", Testing: bool) -> bool: ...
    def saveState(self, Stream: QtCore.QXmlStreamWriter) -> None: ...
    def removeAllDockAreas(self) -> list[CDockAreaWidget]: ...
    def removeDockArea(self, area: typing.Optional["CDockAreaWidget"]) -> None: ...
    def addDockArea(
        self, DockAreaWidget: typing.Optional["CDockAreaWidget"], area: "DockWidgetArea" = ...
    ) -> None: ...
    def dropWidget(
        self,
        Widget: typing.Optional[QtWidgets.QWidget],
        DropArea: "DockWidgetArea",
        TargetAreaWidget: typing.Optional["CDockAreaWidget"],
        TabIndex: int = ...,
    ) -> None: ...
    def dropFloatingWidget(
        self, FloatingWidget: typing.Optional["CFloatingDockContainer"], TargetPos: QtCore.QPoint
    ) -> None: ...
    def createRootSplitter(self) -> None: ...
    def createAndSetupAutoHideContainer(
        self,
        area: "SideBarLocation",
        DockWidget: typing.Optional["CDockWidget"],
        TabIndex: int = ...,
    ) -> typing.Optional["CAutoHideDockContainer"]: ...
    def rootSplitter(self) -> typing.Optional["CDockSplitter"]: ...
    def event(self, e: typing.Optional[QtCore.QEvent]) -> bool: ...

class CDockingStateReader(QtCore.QXmlStreamReader):
    def __init__(self) -> None: ...
    def fileVersion(self) -> int: ...
    def setFileVersion(self, FileVersion: int) -> None: ...

class CDockFocusController(QtCore.QObject):
    def __init__(self, DockManager: typing.Optional["CDockManager"]) -> None: ...
    def setDockWidgetFocused(self, focusedNow: typing.Optional["CDockWidget"]) -> None: ...
    def setDockWidgetTabPressed(self, Value: bool) -> None: ...
    def clearDockWidgetFocus(self, dockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def setDockWidgetTabFocused(self, Tab: typing.Optional["CDockWidgetTab"]) -> None: ...
    def focusedDockWidget(self) -> typing.Optional["CDockWidget"]: ...
    def notifyFloatingWidgetDrop(
        self, FloatingWidget: typing.Optional["CFloatingDockContainer"]
    ) -> None: ...
    def notifyWidgetOrAreaRelocation(
        self, RelocatedWidget: typing.Optional[QtWidgets.QWidget]
    ) -> None: ...

class CDockManager(CDockContainerWidget):
    class eConfigParam(enum.Enum):
        AutoHideOpenOnDragHoverDelay_ms = ...
        ConfigParamCount = ...

    class eAutoHideFlag(enum.Enum):
        AutoHideFeatureEnabled = ...
        DockAreaHasAutoHideButton = ...
        AutoHideButtonTogglesArea = ...
        AutoHideButtonCheckable = ...
        AutoHideSideBarsIconOnly = ...
        AutoHideShowOnMouseOver = ...
        AutoHideCloseButtonCollapsesDock = ...
        AutoHideHasCloseButton = ...
        AutoHideHasMinimizeButton = ...
        AutoHideOpenOnDragHover = ...
        AutoHideCloseOnOutsideMouseClick = ...
        DefaultAutoHideConfig = ...

    class eConfigFlag(enum.Enum):
        ActiveTabHasCloseButton = ...
        DockAreaHasCloseButton = ...
        DockAreaCloseButtonClosesTab = ...
        OpaqueSplitterResize = ...
        XmlAutoFormattingEnabled = ...
        XmlCompressionEnabled = ...
        TabCloseButtonIsToolButton = ...
        AllTabsHaveCloseButton = ...
        RetainTabSizeWhenCloseButtonHidden = ...
        DragPreviewIsDynamic = ...
        DragPreviewShowsContentPixmap = ...
        DragPreviewHasWindowFrame = ...
        AlwaysShowTabs = ...
        DockAreaHasUndockButton = ...
        DockAreaHasTabsMenuButton = ...
        DockAreaHideDisabledButtons = ...
        DockAreaDynamicTabsMenuButtonVisibility = ...
        FloatingContainerHasWidgetTitle = ...
        FloatingContainerHasWidgetIcon = ...
        HideSingleCentralWidgetTitleBar = ...
        FocusHighlighting = ...
        EqualSplitOnInsertion = ...
        FloatingContainerForceNativeTitleBar = ...
        FloatingContainerForceQWidgetTitleBar = ...
        MiddleMouseButtonClosesTab = ...
        DisableTabTextEliding = ...
        ShowTabTextOnlyForActiveTab = ...
        DoubleClickUndocksWidget = ...
        DefaultDockAreaButtons = ...
        DefaultBaseConfig = ...
        DefaultOpaqueConfig = ...
        DefaultNonOpaqueConfig = ...
        NonOpaqueWithWindowFrame = ...

    class eViewMenuInsertionOrder(enum.Enum):
        MenuSortedByInsertion = ...
        MenuAlphabeticallySorted = ...

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = ...) -> None: ...

    focusedDockWidgetChanged: typing.ClassVar[Signal]
    dockWidgetRemoved: typing.ClassVar[Signal]
    dockWidgetAboutToBeRemoved: typing.ClassVar[Signal]
    dockWidgetAdded: typing.ClassVar[Signal]
    dockAreaCreated: typing.ClassVar[Signal]
    floatingWidgetCreated: typing.ClassVar[Signal]
    perspectiveOpened: typing.ClassVar[Signal]
    openingPerspective: typing.ClassVar[Signal]
    stateRestored: typing.ClassVar[Signal]
    restoringState: typing.ClassVar[Signal]
    perspectivesRemoved: typing.ClassVar[Signal]
    perspectiveListLoaded: typing.ClassVar[Signal]
    perspectiveListChanged: typing.ClassVar[Signal]
    def hideManagerAndFloatingWidgets(self) -> None: ...
    def setDockWidgetFocused(self, DockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def openPerspective(self, PerspectiveName: typing.Optional[str]) -> None: ...
    def endLeavingMinimizedState(self) -> None: ...
    def lockDockWidgetFeaturesGlobally(
        self, Features: CDockWidget.DockWidgetFeature = ...
    ) -> None: ...
    def globallyLockedDockWidgetFeatures(self) -> CDockWidget.DockWidgetFeature: ...
    def dockWidgetToolBarIconSize(self, State: CDockWidget.eState) -> QtCore.QSize: ...
    def setDockWidgetToolBarIconSize(
        self, IconSize: QtCore.QSize, State: CDockWidget.eState
    ) -> None: ...
    def dockWidgetToolBarStyle(self, State: CDockWidget.eState) -> QtCore.Qt.ToolButtonStyle: ...
    def setDockWidgetToolBarStyle(
        self, Style: QtCore.Qt.ToolButtonStyle, State: CDockWidget.eState
    ) -> None: ...
    @staticmethod
    def floatingContainersTitle() -> str: ...
    @staticmethod
    def setFloatingContainersTitle(Title: typing.Optional[str]) -> None: ...
    def setSplitterSizes(
        self,
        ContainedArea: typing.Optional["CDockAreaWidget"],
        sizes: collections.abc.Iterable[int],
    ) -> None: ...
    def splitterSizes(self, ContainedArea: typing.Optional["CDockAreaWidget"]) -> list[int]: ...
    def focusedDockWidget(self) -> typing.Optional["CDockWidget"]: ...
    @staticmethod
    def startDragDistance() -> int: ...
    def isLeavingMinimizedState(self) -> bool: ...
    def isRestoringState(self) -> bool: ...
    def setViewMenuInsertionOrder(self, Order: "CDockManager.eViewMenuInsertionOrder") -> None: ...
    def viewMenu(self) -> typing.Optional[QtWidgets.QMenu]: ...
    def addToggleViewActionToMenu(
        self,
        ToggleViewAction: typing.Optional[QtGui.QAction],
        Group: typing.Optional[str] = ...,
        GroupIcon: QtGui.QIcon = ...,
    ) -> typing.Optional[QtGui.QAction]: ...
    def setCentralWidget(
        self, widget: typing.Optional["CDockWidget"]
    ) -> typing.Optional["CDockAreaWidget"]: ...
    def centralWidget(self) -> typing.Optional["CDockWidget"]: ...
    def loadPerspectives(self, Settings: QtCore.QSettings) -> None: ...
    def savePerspectives(self, Settings: QtCore.QSettings) -> None: ...
    def perspectiveNames(self) -> list[str]: ...
    def removePerspectives(self, Names: collections.abc.Iterable[typing.Optional[str]]) -> None: ...
    def removePerspective(self, Name: typing.Optional[str]) -> None: ...
    def addPerspective(self, UniquePrespectiveName: typing.Optional[str]) -> None: ...
    def restoreState(
        self,
        state: typing.Union[QtCore.QByteArray, bytes, bytearray, memoryview],
        version: int = ...,
    ) -> bool: ...
    def saveState(self, version: int = ...) -> QtCore.QByteArray: ...
    def zOrderIndex(self) -> int: ...
    def floatingWidgets(self) -> list["CFloatingDockContainer"]: ...
    def dockContainers(self) -> list["CDockContainerWidget"]: ...
    def dockWidgetsMap(self) -> dict[str, CDockWidget]: ...
    def removeDockWidget(self, Dockwidget: typing.Optional["CDockWidget"]) -> None: ...
    def findDockWidget(
        self, ObjectName: typing.Optional[str]
    ) -> typing.Optional["CDockWidget"]: ...
    def addDockWidgetFloating(
        self, DockWidget: typing.Optional["CDockWidget"]
    ) -> typing.Optional["CFloatingDockContainer"]: ...
    def addDockWidgetTabToArea(
        self,
        Dockwidget: typing.Optional["CDockWidget"],
        DockAreaWidget: typing.Optional["CDockAreaWidget"],
        Index: int = ...,
    ) -> typing.Optional["CDockAreaWidget"]: ...
    def addDockWidgetTab(
        self, area: "DockWidgetArea", Dockwidget: typing.Optional["CDockWidget"]
    ) -> typing.Optional["CDockAreaWidget"]: ...
    def addAutoHideDockWidgetToContainer(
        self,
        Location: "SideBarLocation",
        Dockwidget: typing.Optional["CDockWidget"],
        DockContainerWidget: typing.Optional["CDockContainerWidget"],
    ) -> typing.Optional["CAutoHideDockContainer"]: ...
    def addAutoHideDockWidget(
        self, Location: "SideBarLocation", Dockwidget: typing.Optional["CDockWidget"]
    ) -> typing.Optional["CAutoHideDockContainer"]: ...
    def addDockWidgetToContainer(
        self,
        area: "DockWidgetArea",
        Dockwidget: typing.Optional["CDockWidget"],
        DockContainerWidget: typing.Optional["CDockContainerWidget"] = ...,
    ) -> typing.Optional["CDockAreaWidget"]: ...
    def addDockWidget(
        self,
        area: "DockWidgetArea",
        Dockwidget: typing.Optional["CDockWidget"],
        DockAreaWidget: typing.Optional["CDockAreaWidget"] = ...,
        Index: int = ...,
    ) -> typing.Optional["CDockAreaWidget"]: ...
    @staticmethod
    def iconProvider() -> "CIconProvider": ...
    @staticmethod
    def configParam(Param: "CDockManager.eConfigParam", Default: typing.Any) -> typing.Any: ...
    @staticmethod
    def setConfigParam(Param: "CDockManager.eConfigParam", Value: typing.Any) -> None: ...
    @staticmethod
    def testAutoHideConfigFlag(Flag: "CDockManager.eAutoHideFlag") -> bool: ...
    @staticmethod
    def setAutoHideConfigFlag(Flag: "CDockManager.eAutoHideFlag", On: bool = ...) -> None: ...
    @staticmethod
    def setAutoHideConfigFlags(Flags: "CDockManager.eAutoHideFlag") -> None: ...
    @staticmethod
    def autoHideConfigFlags() -> "CDockManager.eAutoHideFlag": ...
    @staticmethod
    def testConfigFlag(Flag: "CDockManager.eConfigFlag") -> bool: ...
    @staticmethod
    def setConfigFlag(Flag: "CDockManager.eConfigFlag", On: bool = ...) -> None: ...
    @staticmethod
    def setConfigFlags(Flags: "CDockManager.eConfigFlag") -> None: ...
    @staticmethod
    def configFlags() -> "CDockManager.eConfigFlag": ...
    def setComponentsFactory(self, Factory: typing.Optional["CDockComponentsFactory"]) -> None: ...
    def componentsFactory(self) -> typing.Optional["CDockComponentsFactory"]: ...
    def createDockWidget(
        self, title: typing.Optional[str], parent: typing.Optional[QtWidgets.QWidget] = ...
    ) -> typing.Optional["CDockWidget"]: ...
    def showEvent(self, event: typing.Optional[QtGui.QShowEvent]) -> None: ...
    def notifyFloatingWidgetDrop(
        self, FloatingWidget: typing.Optional["CFloatingDockContainer"]
    ) -> None: ...
    def notifyWidgetOrAreaRelocation(
        self, RelocatedWidget: typing.Optional[QtWidgets.QWidget]
    ) -> None: ...
    def dockAreaOverlay(self) -> typing.Optional["CDockOverlay"]: ...
    def containerOverlay(self) -> typing.Optional["CDockOverlay"]: ...
    def removeDockContainer(
        self, DockContainer: typing.Optional["CDockContainerWidget"]
    ) -> None: ...
    def registerDockContainer(
        self, DockContainer: typing.Optional["CDockContainerWidget"]
    ) -> None: ...
    def removeFloatingWidget(
        self, FloatingWidget: typing.Optional["CFloatingDockContainer"]
    ) -> None: ...
    def registerFloatingWidget(
        self, FloatingWidget: typing.Optional["CFloatingDockContainer"]
    ) -> None: ...

class CDockOverlay(QtWidgets.QFrame):
    class eMode(enum.Enum):
        ModeDockAreaOverlay = ...
        ModeContainerOverlay = ...

    def __init__(
        self, parent: typing.Optional[QtWidgets.QWidget], Mode: "CDockOverlay.eMode" = ...
    ) -> None: ...
    def hideEvent(self, e: typing.Optional[QtGui.QHideEvent]) -> None: ...
    def showEvent(self, e: typing.Optional[QtGui.QShowEvent]) -> None: ...
    def paintEvent(self, e: typing.Optional[QtGui.QPaintEvent]) -> None: ...
    def event(self, e: typing.Optional[QtCore.QEvent]) -> bool: ...
    def dropOverlayRect(self) -> QtCore.QRect: ...
    def dropPreviewEnabled(self) -> bool: ...
    def enableDropPreview(self, Enable: bool) -> None: ...
    def hideOverlay(self) -> None: ...
    def showOverlay(self, target: typing.Optional[QtWidgets.QWidget]) -> "DockWidgetArea": ...
    def visibleDropAreaUnderCursor(self) -> "DockWidgetArea": ...
    def tabIndexUnderCursor(self) -> int: ...
    def dropAreaUnderCursor(self) -> "DockWidgetArea": ...
    def allowedAreas(self) -> "DockWidgetArea": ...
    def setAllowedArea(self, area: "DockWidgetArea", Enable: bool) -> None: ...
    def setAllowedAreas(self, areas: "DockWidgetArea") -> None: ...

class CDockOverlayCross(QtWidgets.QWidget):
    class eIconColor(enum.Enum):
        FrameColor = ...
        WindowBackgroundColor = ...
        OverlayColor = ...
        ArrowColor = ...
        ShadowColor = ...

    def __init__(self, overlay: typing.Optional["CDockOverlay"]) -> None: ...
    def setIconColors(self, Colors: typing.Optional[str]) -> None: ...
    def updatePosition(self) -> None: ...
    def reset(self) -> None: ...
    def updateOverlayIcons(self) -> None: ...
    def setupOverlayCross(self, Mode: "CDockOverlay.eMode") -> None: ...
    def cursorLocation(self) -> "DockWidgetArea": ...
    def setIconColor(
        self,
        ColorIndex: "CDockOverlayCross.eIconColor",
        Color: typing.Union[QtGui.QColor, QtCore.Qt.GlobalColor, int],
    ) -> None: ...
    def setAreaWidgets(self, widgets: dict["DockWidgetArea", QtWidgets.QWidget]) -> None: ...
    def showEvent(self, e: typing.Optional[QtGui.QShowEvent]) -> None: ...
    def setIconShadowColor(
        self, Color: typing.Union[QtGui.QColor, QtCore.Qt.GlobalColor, int]
    ) -> None: ...
    def setIconArrowColor(
        self, Color: typing.Union[QtGui.QColor, QtCore.Qt.GlobalColor, int]
    ) -> None: ...
    def setIconOverlayColor(
        self, Color: typing.Union[QtGui.QColor, QtCore.Qt.GlobalColor, int]
    ) -> None: ...
    def setIconBackgroundColor(
        self, Color: typing.Union[QtGui.QColor, QtCore.Qt.GlobalColor, int]
    ) -> None: ...
    def setIconFrameColor(
        self, Color: typing.Union[QtGui.QColor, QtCore.Qt.GlobalColor, int]
    ) -> None: ...
    @typing.overload
    def iconColor(self) -> QtGui.QColor: ...
    @typing.overload
    def iconColor(self, ColorIndex: "CDockOverlayCross.eIconColor") -> QtGui.QColor: ...
    def iconColors(self) -> str: ...

class CDockSplitter(QtWidgets.QSplitter):
    @typing.overload
    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = ...) -> None: ...
    @typing.overload
    def __init__(
        self, orientation: QtCore.Qt.Orientation, parent: typing.Optional[QtWidgets.QWidget] = ...
    ) -> None: ...
    def isResizingWithContainer(self) -> bool: ...
    def lastWidget(self) -> typing.Optional[QtWidgets.QWidget]: ...
    def firstWidget(self) -> typing.Optional[QtWidgets.QWidget]: ...
    def hasVisibleContent(self) -> bool: ...

class CDockWidgetTab(QtWidgets.QFrame):
    def __init__(
        self,
        DockWidget: typing.Optional["CDockWidget"],
        parent: typing.Optional[QtWidgets.QWidget] = ...,
    ) -> None: ...

    elidedChanged: typing.ClassVar[Signal]
    moved: typing.ClassVar[Signal]
    closeOtherTabsRequested: typing.ClassVar[Signal]
    closeRequested: typing.ClassVar[Signal]
    clicked: typing.ClassVar[Signal]
    activeTabChanged: typing.ClassVar[Signal]
    def setVisible(self, visible: bool) -> None: ...
    def buildContextMenu(
        self, menu: typing.Optional[QtWidgets.QMenu] = ...
    ) -> typing.Optional[QtWidgets.QMenu]: ...
    def dragState(self) -> "eDragState": ...
    def setIconSize(self, Size: QtCore.QSize) -> None: ...
    def iconSize(self) -> QtCore.QSize: ...
    def updateStyle(self) -> None: ...
    def setElideMode(self, mode: QtCore.Qt.TextElideMode) -> None: ...
    def event(self, e: typing.Optional[QtCore.QEvent]) -> bool: ...
    def isClosable(self) -> bool: ...
    def isTitleElided(self) -> bool: ...
    def setText(self, title: typing.Optional[str]) -> None: ...
    def text(self) -> str: ...
    def icon(self) -> QtGui.QIcon: ...
    def setIcon(self, Icon: QtGui.QIcon) -> None: ...
    def dockWidget(self) -> typing.Optional["CDockWidget"]: ...
    def dockAreaWidget(self) -> typing.Optional["CDockAreaWidget"]: ...
    def setDockAreaWidget(self, DockArea: typing.Optional["CDockAreaWidget"]) -> None: ...
    def setActiveTab(self, active: bool) -> None: ...
    def isActiveTab(self) -> bool: ...
    def mouseDoubleClickEvent(self, event: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def contextMenuEvent(self, ev: typing.Optional[QtGui.QContextMenuEvent]) -> None: ...
    def mouseMoveEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mouseReleaseEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mousePressEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...

class CElidingLabel(QtWidgets.QLabel):
    @typing.overload
    def __init__(
        self, parent: typing.Optional[QtWidgets.QWidget] = ..., f: QtCore.Qt.WindowType = ...
    ) -> None: ...
    @typing.overload
    def __init__(
        self,
        text: typing.Optional[str],
        parent: typing.Optional[QtWidgets.QWidget] = ...,
        f: QtCore.Qt.WindowType = ...,
    ) -> None: ...

    elidedChanged: typing.ClassVar[Signal]
    doubleClicked: typing.ClassVar[Signal]
    clicked: typing.ClassVar[Signal]
    def text(self) -> str: ...
    def setText(self, text: typing.Optional[str]) -> None: ...
    def sizeHint(self) -> QtCore.QSize: ...
    def minimumSizeHint(self) -> QtCore.QSize: ...
    def isElided(self) -> bool: ...
    def setElideMode(self, mode: QtCore.Qt.TextElideMode) -> None: ...
    def elideMode(self) -> QtCore.Qt.TextElideMode: ...
    def mouseDoubleClickEvent(self, ev: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def resizeEvent(self, event: typing.Optional[QtGui.QResizeEvent]) -> None: ...
    def mouseReleaseEvent(self, event: typing.Optional[QtGui.QMouseEvent]) -> None: ...

class IFloatingWidget:
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: "IFloatingWidget") -> None: ...
    def finishDragging(self) -> None: ...
    def moveFloating(self) -> None: ...
    def startFloating(
        self,
        DragStartMousePos: QtCore.QPoint,
        Size: QtCore.QSize,
        DragState: "eDragState",
        MouseEventHandler: typing.Optional[QtWidgets.QWidget],
    ) -> None: ...

class CFloatingDockContainer(QtWidgets.QWidget, IFloatingWidget):
    @typing.overload
    def __init__(self, DockManager: typing.Optional["CDockManager"]) -> None: ...
    @typing.overload
    def __init__(self, DockArea: typing.Optional["CDockAreaWidget"]) -> None: ...
    @typing.overload
    def __init__(self, DockWidget: typing.Optional["CDockWidget"]) -> None: ...
    def finishDropOperation(self) -> None: ...
    def dockWidgets(self) -> list["CDockWidget"]: ...
    def topLevelDockWidget(self) -> typing.Optional["CDockWidget"]: ...
    def hasTopLevelDockWidget(self) -> bool: ...
    def isClosable(self) -> bool: ...
    def startDragging(
        self,
        DragStartMousePos: QtCore.QPoint,
        Size: QtCore.QSize,
        MouseEventHandler: typing.Optional[QtWidgets.QWidget],
    ) -> None: ...
    def dockContainer(self) -> typing.Optional["CDockContainerWidget"]: ...
    def moveEvent(self, event: typing.Optional[QtGui.QMoveEvent]) -> None: ...
    def event(self, e: typing.Optional[QtCore.QEvent]) -> bool: ...
    def showEvent(self, event: typing.Optional[QtGui.QShowEvent]) -> None: ...
    def hideEvent(self, event: typing.Optional[QtGui.QHideEvent]) -> None: ...
    def closeEvent(self, event: typing.Optional[QtGui.QCloseEvent]) -> None: ...
    def changeEvent(self, event: typing.Optional[QtCore.QEvent]) -> None: ...
    def updateWindowTitle(self) -> None: ...
    def restoreState(self, Stream: "CDockingStateReader", Testing: bool) -> bool: ...
    def moveFloating(self) -> None: ...
    def initFloatingGeometry(
        self, DragStartMousePos: QtCore.QPoint, Size: QtCore.QSize
    ) -> None: ...
    def deleteContent(self) -> None: ...
    def finishDragging(self) -> None: ...
    def startFloating(
        self,
        DragStartMousePos: QtCore.QPoint,
        Size: QtCore.QSize,
        DragState: "eDragState",
        MouseEventHandler: typing.Optional[QtWidgets.QWidget],
    ) -> None: ...

class CFloatingDragPreview(QtWidgets.QWidget, IFloatingWidget):
    @typing.overload
    def __init__(
        self,
        Content: typing.Optional[QtWidgets.QWidget],
        parent: typing.Optional[QtWidgets.QWidget],
    ) -> None: ...
    @typing.overload
    def __init__(self, Content: typing.Optional["CDockWidget"]) -> None: ...
    @typing.overload
    def __init__(self, Content: typing.Optional["CDockAreaWidget"]) -> None: ...

    draggingCanceled: typing.ClassVar[Signal]
    def cleanupAutoHideContainerWidget(self, ContainerDropArea: "DockWidgetArea") -> None: ...
    def finishDragging(self) -> None: ...
    def moveFloating(self) -> None: ...
    def startFloating(
        self,
        DragStartMousePos: QtCore.QPoint,
        Size: QtCore.QSize,
        DragState: "eDragState",
        MouseEventHandler: typing.Optional[QtWidgets.QWidget],
    ) -> None: ...
    def eventFilter(
        self, watched: typing.Optional[QtCore.QObject], event: typing.Optional[QtCore.QEvent]
    ) -> bool: ...
    def paintEvent(self, e: typing.Optional[QtGui.QPaintEvent]) -> None: ...

class CIconProvider:
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: CIconProvider) -> None: ...
    def registerCustomIcon(self, IconId: "eIcon", icon: QtGui.QIcon) -> None: ...
    def customIcon(self, IconId: "eIcon") -> QtGui.QIcon: ...

class CResizeHandle(QtWidgets.QFrame):
    def __init__(
        self, HandlePosition: QtCore.Qt.Edge, parent: typing.Optional[QtWidgets.QWidget]
    ) -> None: ...
    def opaqueResize(self) -> bool: ...
    def setOpaqueResize(self, opaque: bool = ...) -> None: ...
    def setMaxResizeSize(self, MaxSize: int) -> None: ...
    def setMinResizeSize(self, MinSize: int) -> None: ...
    def isResizing(self) -> bool: ...
    def sizeHint(self) -> QtCore.QSize: ...
    def orientation(self) -> QtCore.Qt.Orientation: ...
    def handlePostion(self) -> QtCore.Qt.Edge: ...
    def setHandlePosition(self, HandlePosition: QtCore.Qt.Edge) -> None: ...
    def mouseReleaseEvent(self, a0: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mousePressEvent(self, a0: typing.Optional[QtGui.QMouseEvent]) -> None: ...
    def mouseMoveEvent(self, a0: typing.Optional[QtGui.QMouseEvent]) -> None: ...



from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from pydantic import BaseModel, Field
from qtpy.QtCore import QObject, Signal


@dataclass
class ImageStats:
    """Container to store stats of an image."""

    maximum: float
    minimum: float
    mean: float
    std: float

    @classmethod
    def from_data(cls, data: np.ndarray) -> ImageStats:
        """
        Get the statistics of the image data.

        Args:
            data(np.ndarray): The image data.

        Returns:
            ImageStats: The statistics of the image data.
        """
        return cls(
            maximum=np.nanmax(data),
            minimum=np.nanmin(data),
            mean=np.nanmean(data),
            std=np.nanstd(data),
        )


# noinspection PyDataclass
class ProcessingConfig(BaseModel):
    fft: bool = Field(False, description="Whether to perform FFT on the monitor data.")
    log: bool = Field(False, description="Whether to perform log on the monitor data.")
    transpose: bool = Field(
        False, description="Whether to transpose the monitor data before displaying."
    )
    num_rotation_90: int = Field(
        0, description="The rotation angle of the monitor data before displaying."
    )
    stats: ImageStats = Field(
        ImageStats(maximum=0, minimum=0, mean=0, std=0),
        description="The statistics of the image data.",
    )

    model_config: dict = {"validate_assignment": True}


class ImageProcessor(QObject):
    """
    Class for processing the image data.
    """

    image_processed = Signal(np.ndarray)

    def __init__(self, parent=None, config: ProcessingConfig = None):
        super().__init__(parent=parent)
        if config is None:
            config = ProcessingConfig()
        self.config = config
        self._current_thread = None

    def set_config(self, config: ProcessingConfig):
        """
        Set the configuration of the processor.

        Args:
            config(ProcessingConfig): The configuration of the processor.
        """
        self.config = config

    def FFT(self, data: np.ndarray) -> np.ndarray:
        """
        Perform FFT on the data.

        Args:
            data(np.ndarray): The data to be processed.

        Returns:
            np.ndarray: The processed data.
        """
        return np.abs(np.fft.fftshift(np.fft.fft2(np.nan_to_num(data))))

    def rotation(self, data: np.ndarray, rotate_90: int) -> np.ndarray:
        """
        Rotate the data by 90 degrees n times.

        Args:
            data(np.ndarray): The data to be processed.
            rotate_90(int): The number of 90 degree rotations.

        Returns:
            np.ndarray: The processed data.
        """
        return np.rot90(data, k=rotate_90, axes=(0, 1))

    def transpose(self, data: np.ndarray) -> np.ndarray:
        """
        Transpose the data.

        Args:
            data(np.ndarray): The data to be processed.

        Returns:
            np.ndarray: The processed data.
        """
        return np.transpose(data)

    def log(self, data: np.ndarray) -> np.ndarray:
        """
        Perform log on the data.

        Args:
            data(np.ndarray): The data to be processed.

        Returns:
            np.ndarray: The processed data.
        """
        # TODO this is not final solution -> data should stay as int16
        data = data.astype(np.float32)
        offset = 1e-6
        data_offset = data + offset
        return np.log10(data_offset)

    def update_image_stats(self, data: np.ndarray) -> None:
        """Get the statistics of the image data.

        Args:
            data(np.ndarray): The image data.

        """
        self.config.stats.maximum = np.max(data)
        self.config.stats.minimum = np.min(data)
        self.config.stats.mean = np.mean(data)
        self.config.stats.std = np.std(data)

    def process_image(self, data: np.ndarray) -> np.ndarray:
        """Core processing logic without threading overhead."""
        if self.config.fft:
            data = self.FFT(data)
        if self.config.num_rotation_90 is not None:
            data = self.rotation(data, self.config.num_rotation_90)
        if self.config.transpose:
            data = self.transpose(data)
        if self.config.log:
            data = self.log(data)
        self.update_image_stats(data)
        return data



"""Client utilities for the BEC GUI."""

from __future__ import annotations

import json
import os
import select
import subprocess
import threading
import time
from contextlib import contextmanager
from threading import Lock
from typing import TYPE_CHECKING, Literal, TypeAlias, cast

from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.utils.import_utils import lazy_import, lazy_import_from
from rich.console import Console
from rich.table import Table

from bec_widgets.cli.rpc.rpc_base import RPCBase, RPCReference
from bec_widgets.utils.serialization import register_serializer_extension

if TYPE_CHECKING:  # pragma: no cover
    from bec_lib.messages import GUIRegistryStateMessage

    import bec_widgets.cli.client as client
else:
    GUIRegistryStateMessage = lazy_import_from("bec_lib.messages", "GUIRegistryStateMessage")
    client = lazy_import("bec_widgets.cli.client")


logger = bec_logger.logger

IGNORE_WIDGETS = ["LaunchWindow"]

RegistryState: TypeAlias = dict[
    Literal["gui_id", "name", "widget_class", "config", "__rpc__", "container_proxy"],
    str | bool | dict,
]

# pylint: disable=redefined-outer-scope


def _filter_output(output: str) -> str:
    """
    Filter out the output from the process.
    """
    if "IMKClient" in output:
        # only relevant on macOS
        # see https://discussions.apple.com/thread/255761734?sortBy=rank
        return ""
    return output


def _get_output(process, logger) -> None:
    log_func = {process.stdout: logger.debug, process.stderr: logger.info}
    stream_buffer = {process.stdout: [], process.stderr: []}
    try:
        os.set_blocking(process.stdout.fileno(), False)
        os.set_blocking(process.stderr.fileno(), False)
        while process.poll() is None:
            readylist, _, _ = select.select([process.stdout, process.stderr], [], [], 1)
            for stream in (process.stdout, process.stderr):
                buf = stream_buffer[stream]
                if stream in readylist:
                    buf.append(stream.read(4096))
                output, _, remaining = "".join(buf).rpartition("\n")
                output = _filter_output(output)
                if output:
                    log_func[stream](output)
                    buf.clear()
                    buf.append(remaining)
    except Exception as e:
        logger.error(f"Error reading process output: {str(e)}")


def _start_plot_process(
    gui_id: str,
    gui_class_id: str,
    config: dict | str,
    gui_class: str = "dock_area",
    logger=None,  # FIXME change gui_class back to "launcher" later
) -> tuple[subprocess.Popen[str], threading.Thread | None]:
    """
    Start the plot in a new process.

    Logger must be a logger object with "debug" and "error" functions,
    or it can be left to "None" as default. None means output from the
    process will not be captured.
    """
    # pylint: disable=subprocess-run-check
    command = [
        "bec-gui-server",
        "--id",
        gui_id,
        "--gui_class",
        gui_class,
        "--gui_class_id",
        gui_class_id,
        "--hide",
    ]
    if config:
        if isinstance(config, dict):
            config = json.dumps(config)
        command.extend(["--config", str(config)])

    env_dict = os.environ.copy()
    env_dict["PYTHONUNBUFFERED"] = "1"

    if logger is None:
        stdout_redirect = subprocess.DEVNULL
        stderr_redirect = subprocess.DEVNULL
    else:
        stdout_redirect = subprocess.PIPE
        stderr_redirect = subprocess.PIPE

    process = subprocess.Popen(
        command,
        text=True,
        start_new_session=True,
        stdout=stdout_redirect,
        stderr=stderr_redirect,
        env=env_dict,
    )
    if logger is None:
        process_output_processing_thread = None
    else:
        process_output_processing_thread = threading.Thread(
            target=_get_output, args=(process, logger)
        )
        process_output_processing_thread.start()
    return process, process_output_processing_thread


class RepeatTimer(threading.Timer):
    """RepeatTimer class."""

    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)


# pylint: disable=protected-access
@contextmanager
def wait_for_server(client: BECGuiClient):
    """Context manager to wait for the server to start."""
    timeout = client._startup_timeout
    if not timeout:
        if client._gui_is_alive():
            # there is hope, let's wait a bit
            timeout = 1
        else:
            raise RuntimeError("GUI is not alive")
    try:
        if client._gui_started_event.wait(timeout=timeout):
            if client._gui_started_timer is not None:
                # cancel the timer, we are done
                client._gui_started_timer.cancel()
                client._gui_started_timer.join()
        else:
            raise TimeoutError("Could not connect to GUI server")
    finally:
        # after initial waiting period, do not wait so much any more
        # (only relevant if GUI didn't start)
        client._startup_timeout = 0
    yield


class WidgetNameSpace:
    def __repr__(self):
        console = Console()
        table = Table(title="Available widgets for BEC CLI usage")
        table.add_column("Widget Name", justify="left", style="magenta")
        table.add_column("Description", justify="left")
        for attr, value in self.__dict__.items():
            docs = value.__doc__
            docs = docs if docs else "No description available"
            table.add_row(attr, docs)
        console.print(table)
        return ""


class AvailableWidgetsNamespace:
    """Namespace for available widgets in the BEC GUI."""

    def __init__(self):
        for widget in client.Widgets:
            name = widget.value
            if name in IGNORE_WIDGETS:
                continue
            setattr(self, name, name)

    def __repr__(self):
        console = Console()
        table = Table(title="Available widgets for BEC CLI usage")
        table.add_column("Widget Name", justify="left", style="magenta")
        table.add_column("Description", justify="left")
        for attr_name, _ in self.__dict__.items():
            docs = getattr(client, attr_name).__doc__
            docs = docs if docs else "No description available"
            table.add_row(attr_name, docs if len(docs.strip()) > 0 else "No description available")
        console.print(table)
        return ""


class BECGuiClient(RPCBase):
    """BEC GUI client class. Container for GUI applications within Python."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._lock = Lock()
        self._anchor_widget = "launcher"
        self._killed = False
        self._top_level: dict[str, RPCReference] = {}
        self._startup_timeout = 0
        self._gui_started_timer = None
        self._gui_started_event = threading.Event()
        self._process = None
        self._process_output_processing_thread = None
        self._server_registry: dict[str, RegistryState] = {}
        self._ipython_registry: dict[str, RPCReference] = {}
        self.available_widgets = AvailableWidgetsNamespace()
        register_serializer_extension()

    ####################
    #### Client API ####
    ####################

    @property
    def launcher(self) -> RPCBase:
        """The launcher object."""
        return RPCBase(gui_id=f"{self._gui_id}:launcher", parent=self, object_name="launcher")

    def connect_to_gui_server(self, gui_id: str) -> None:
        """Connect to a GUI server"""
        # Unregister the old callback
        self._client.connector.unregister(
            MessageEndpoints.gui_registry_state(self._gui_id), cb=self._handle_registry_update
        )
        self._gui_id = gui_id

        # reset the namespace
        self._update_dynamic_namespace({})
        self._server_registry = {}
        self._top_level = {}
        self._ipython_registry = {}

        # Register the new callback
        self._client.connector.register(
            MessageEndpoints.gui_registry_state(self._gui_id),
            cb=self._handle_registry_update,
            parent=self,
            from_start=True,
        )

    @property
    def windows(self) -> dict:
        """Dictionary with dock areas in the GUI."""
        return {widget.object_name: widget for widget in self._top_level.values()}

    @property
    def window_list(self) -> list:
        """List with dock areas in the GUI."""
        return list(self._top_level.values())

    def start(self, wait: bool = False) -> None:
        """Start the GUI server."""
        logger.warning("Using <gui>.start() is deprecated, use <gui>.show() instead.")
        return self._start(wait=wait)

    def show(self, wait=True) -> None:
        """
        Show the GUI window.
        If the GUI server is not running, it will be started.

        Args:
            wait(bool): Whether to wait for the server to start. Defaults to True.
        """
        if self._check_if_server_is_alive():
            return self._show_all()
        return self._start(wait=wait)

    def hide(self):
        """Hide the GUI window."""
        return self._hide_all()

    def raise_window(self, wait: bool = True) -> None:
        """
        Bring GUI windows to the front.
        If the GUI server is not running, it will be started.

        Args:
            wait(bool): Whether to wait for the server to start. Defaults to True.
        """
        if self._check_if_server_is_alive():
            return self._raise_all()
        return self._start(wait=wait)

    def change_theme(self, theme: Literal["light", "dark"] | None = None) -> None:
        """
        Apply a GUI theme or toggle between dark and light.

        Args:
            theme(Literal["light", "dark"] | None): Theme to apply. If None, the current
                theme is fetched from the GUI and toggled.
        """
        if not self._check_if_server_is_alive():
            self._start(wait=True)

        with wait_for_server(self):
            if theme is None:
                current_theme = self.launcher._run_rpc("fetch_theme")
                next_theme = "light" if current_theme == "dark" else "dark"
            else:
                next_theme = theme
            self.launcher._run_rpc("change_theme", theme=next_theme)

    def new(
        self,
        name: str | None = None,
        wait: bool = True,
        geometry: tuple[int, int, int, int] | None = None,
        launch_script: str = "dock_area",
        startup_profile: str | Literal["restore", "skip"] | None = None,
        **kwargs,
    ) -> client.AdvancedDockArea:
        """Create a new top-level dock area.

        Args:
            name(str, optional): The name of the dock area. Defaults to None.
            wait(bool, optional): Whether to wait for the server to start. Defaults to True.
            geometry(tuple[int, int, int, int] | None): The geometry of the dock area (pos_x, pos_y, w, h).
            launch_script(str): The launch script to use. Defaults to "dock_area".
            startup_profile(str | Literal["restore", "skip"] | None): Startup mode for
                the dock area:
                  - None: start in transient empty workspace
                  - "restore": restore last-used profile
                  - "skip": skip profile initialization
                  - "<name>": load the named profile
            **kwargs: Additional keyword arguments passed to the dock area.

        Returns:
            client.AdvancedDockArea: The new dock area.

        Examples:
            >>> gui.new()  # Start with an empty unsaved workspace
            >>> gui.new(startup_profile="restore")  # Restore last profile
            >>> gui.new(startup_profile="my_profile")  # Load explicit profile
        """
        if "profile" in kwargs or "start_empty" in kwargs:
            raise TypeError(
                "gui.new() no longer accepts 'profile' or 'start_empty'. Use 'startup_profile' instead."
            )

        if not self._check_if_server_is_alive():
            self.start(wait=True)
        if wait:
            with wait_for_server(self):
                return self._new_impl(
                    name=name,
                    geometry=geometry,
                    launch_script=launch_script,
                    startup_profile=startup_profile,
                    **kwargs,
                )
        return self._new_impl(
            name=name,
            geometry=geometry,
            launch_script=launch_script,
            startup_profile=startup_profile,
            **kwargs,
        )

    def _new_impl(
        self,
        *,
        name: str | None,
        geometry: tuple[int, int, int, int] | None,
        launch_script: str,
        startup_profile: str | Literal["restore", "skip"] | None,
        **kwargs,
    ):
        if launch_script == "dock_area":
            try:
                return self.launcher._run_rpc(
                    "system.launch_dock_area",
                    name=name,
                    geometry=geometry,
                    startup_profile=startup_profile,
                    **kwargs,
                )
            except ValueError as exc:
                error = str(exc)
                if (
                    "Unknown system RPC method: system.launch_dock_area" not in error
                    and "has no attribute 'system.launch_dock_area'" not in error
                ):
                    raise
                logger.debug("Server does not support system.launch_dock_area; using launcher RPC")

        return self.launcher._run_rpc(
            "launch",
            launch_script=launch_script,
            name=name,
            geometry=geometry,
            startup_profile=startup_profile,
            **kwargs,
        )  # pylint: disable=protected-access

    def delete(self, name: str) -> None:
        """Delete a dock area and its parent window.

        Args:
            name(str): The name of the dock area.
        """
        widget = self.windows.get(name)
        if widget is None:
            raise ValueError(f"Dock area {name} not found.")

        # Get the container_proxy (parent window) gui_id from the server registry
        obj = self._server_registry.get(widget._gui_id)
        if obj is None:
            raise ValueError(f"Widget {name} not found in registry.")

        container_gui_id = obj.get("container_proxy")
        if container_gui_id:
            # Close the container window which will also clean up the dock area
            widget._run_rpc("close", gui_id=container_gui_id)  # pylint: disable=protected-access
        else:
            # Fallback: just close the dock area directly
            widget._run_rpc("close")  # pylint: disable=protected-access

    def delete_all(self) -> None:
        """Delete all dock areas."""
        for widget_name in self.windows:
            self.delete(widget_name)

    def kill_server(self) -> None:
        """Kill the GUI server."""
        # Unregister the registry state
        self._killed = True

        if self._gui_started_timer is not None:
            self._gui_started_timer.cancel()
            self._gui_started_timer.join()

        if self._process is None:
            return

        if self._process:
            logger.success("Stopping GUI...")
            self._process.terminate()
            if self._process_output_processing_thread:
                self._process_output_processing_thread.join()
            self._process.wait()
            self._process = None

        # Unregister the registry state
        self._client.connector.unregister(
            MessageEndpoints.gui_registry_state(self._gui_id), cb=self._handle_registry_update
        )
        # Remove all reference from top level
        self._top_level.clear()
        self._server_registry.clear()

    def close(self):
        """Deprecated. Use kill_server() instead."""
        # FIXME, deprecated in favor of kill, will be removed in the future
        self.kill_server()

    #########################
    #### Private methods ####
    #########################

    def _check_if_server_is_alive(self):
        """Checks if the process is alive"""
        if self._process is None:
            return False
        if self._process.poll() is not None:
            return False
        return True

    def _gui_post_startup(self):
        timeout = 60
        # Wait for 'bec' gui to be registered, this may take some time
        # After 60s timeout. Should this raise an exception on timeout?
        start = time.monotonic()
        while time.monotonic() < start + timeout:
            if len(list(self._server_registry.keys())) < 2 or not hasattr(
                self, self._anchor_widget
            ):
                time.sleep(0.1)
            else:
                break

        self._gui_started_event.set()

    def _start_server(self, wait: bool = False) -> None:
        """
        Start the GUI server, and execute callback when it is launched
        """
        if self._gui_is_alive():
            self._gui_started_event.set()
            return
        if self._process is None or self._process.poll() is not None:
            logger.success("GUI starting...")
            self._startup_timeout = 5
            self._gui_started_event.clear()
            self._process, self._process_output_processing_thread = _start_plot_process(
                self._gui_id,
                gui_class_id="bec",
                config=self._client._service_config.config,  # pylint: disable=protected-access
                logger=logger,
            )

            def gui_started_callback(callback):
                try:
                    if callable(callback):
                        callback()
                finally:
                    threading.current_thread().cancel()  # type: ignore

            self._gui_started_timer = RepeatTimer(
                0.5, lambda: self._gui_is_alive() and gui_started_callback(self._gui_post_startup)
            )
            self._gui_started_timer.start()

        if wait:
            self._gui_started_event.wait()

    def _start(self, wait: bool = False) -> None:
        self._killed = False
        self._client.connector.register(
            MessageEndpoints.gui_registry_state(self._gui_id),
            cb=self._handle_registry_update,
            parent=self,
        )
        return self._start_server(wait=wait)

    @staticmethod
    def _handle_registry_update(
        msg: dict[str, GUIRegistryStateMessage], parent: BECGuiClient
    ) -> None:
        # This was causing a deadlock during shutdown, not sure why.
        # with self._lock:
        self = parent
        self._server_registry = cast(dict[str, RegistryState], msg["data"].state)
        self._update_dynamic_namespace(self._server_registry)

    def _do_show_all(self):
        if self.launcher and len(self._top_level) == 0:
            self.launcher._run_rpc("show")  # pylint: disable=protected-access
        for window in self._top_level.values():
            window.show()

    def _show_all(self):
        with wait_for_server(self):
            return self._do_show_all()

    def _hide_all(self):
        with wait_for_server(self):
            if self._killed:
                return
            self.launcher._run_rpc("hide")
            for window in self._top_level.values():
                window.hide()

    def _do_raise_all(self):
        """Bring GUI windows to the front."""
        if self.launcher and len(self._top_level) == 0:
            self.launcher._run_rpc("raise")  # pylint: disable=protected-access
        for window in self._top_level.values():
            window._run_rpc("raise")  # type: ignore[attr-defined]

    def _raise_all(self):
        with wait_for_server(self):
            if self._killed:
                return
            return self._do_raise_all()

    def _update_dynamic_namespace(self, server_registry: dict):
        """
        Update the dynamic name space with the given server registry.
        Setting the server registry to an empty dictionary will remove all widgets from the namespace.

        Args:
            server_registry (dict): The server registry
        """
        top_level_widgets: dict[str, RPCReference] = {}
        for gui_id, state in server_registry.items():
            widget = self._add_widget(state, self)
            if widget is None:
                # ignore widgets that are not supported
                continue
            # get all top-level widgets. These are widgets that have no parent
            if not state["config"].get("parent_id"):
                top_level_widgets[gui_id] = widget

        remove_from_registry = []
        for gui_id, widget in self._ipython_registry.items():
            if gui_id not in server_registry:
                remove_from_registry.append(gui_id)
        for gui_id in remove_from_registry:
            self._ipython_registry.pop(gui_id)

        removed_widgets = [
            widget.object_name for widget in self._top_level.values() if widget._is_deleted()
        ]

        for widget_name in removed_widgets:
            # the check is not strictly necessary, but better safe
            # than sorry; who knows what the user has done
            if hasattr(self, widget_name):
                delattr(self, widget_name)

        for gui_id, widget_ref in top_level_widgets.items():
            setattr(self, widget_ref.object_name, widget_ref)

        self._top_level = top_level_widgets

        for widget in self._ipython_registry.values():
            widget._refresh_references()

    def _add_widget(self, state: dict, parent: object) -> RPCReference | None:
        """Add a widget to the namespace

        Args:
            state (dict): The state of the widget from the _server_registry.
            parent (object): The parent object.
        """
        object_name = state["object_name"]
        gui_id = state["gui_id"]
        if state["widget_class"] in IGNORE_WIDGETS:
            return
        widget_class = getattr(client, state["widget_class"], None)
        if widget_class is None:
            return
        obj = self._ipython_registry.get(gui_id)
        if obj is None:
            widget = widget_class(gui_id=gui_id, object_name=object_name, parent=parent)
            self._ipython_registry[gui_id] = widget
        else:
            widget = obj
        obj = RPCReference(registry=self._ipython_registry, gui_id=gui_id)
        return obj


if __name__ == "__main__":  # pragma: no cover
    from bec_lib.client import BECClient
    from bec_lib.service_config import ServiceConfig

    try:
        config = ServiceConfig()
        bec_client = BECClient(config)
        bec_client.start()

        # Test the client_utils.py module
        gui = BECGuiClient()

        gui.show(wait=True)
        gui.new().new(widget="Waveform")
        time.sleep(10)
    finally:
        gui.kill_server()



# Copyright (C) 2022 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR BSD-3-Clause
from qtpy.QtCore import QSize
from qtpy.QtDesigner import QDesignerCustomWidgetInterface
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QWidget

from bec_widgets.utils.bec_designer import designer_material_icon
from bec_widgets.widgets.containers.main_window.main_window import BECMainWindow

DOM_XML = """
<ui language='c++'>
    <widget class='BECMainWindow' name='bec_main_window'>
    </widget>
</ui>
"""


class BECMainWindowPlugin(QDesignerCustomWidgetInterface):  # pragma: no cover
    def __init__(self):
        super().__init__()
        self._form_editor = None

    def createWidget(self, parent):
        # We want to initialize BECMainWindow upon starting designer
        t = BECMainWindow(parent)
        return t

    def domXml(self):
        return DOM_XML

    def group(self):
        return "BEC Containers"

    def icon(self):
        return designer_material_icon(BECMainWindow.ICON_NAME)

    def includeFile(self):
        return "bec_main_window"

    def initialize(self, form_editor):
        import os

        from qtpy.QtCore import Qt
        from qtpy.QtWidgets import QApplication

        import bec_widgets

        MODULE_PATH = os.path.dirname(bec_widgets.__file__)
        QApplication.setAttribute(Qt.AA_DontUseNativeMenuBar, True)
        app = QApplication.instance()
        icon = QIcon()
        icon.addFile(
            os.path.join(MODULE_PATH, "assets", "app_icons", "BEC-General-App.png"),
            size=QSize(48, 48),
        )
        app.setWindowIcon(icon)
        self._form_editor = form_editor

    def isContainer(self):
        return True

    def isInitialized(self):
        return self._form_editor is not None

    def name(self):
        return "BECMainWindow"

    def toolTip(self):
        return "BECMainWindow"

    def whatsThis(self):
        return self.toolTip()



from bec_lib.device import Device
from bec_qthemes import material_icon
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QHBoxLayout, QLabel, QToolButton, QVBoxLayout, QWidget

from bec_widgets.utils.bec_connector import ConnectionConfig
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.utils.ophyd_kind_util import Kind
from bec_widgets.widgets.utility.signal_label.signal_label import SignalLabel


class SignalDisplay(BECWidget, QWidget):
    RPC = False

    def __init__(
        self,
        parent=None,
        client=None,
        device: str = "",
        config: ConnectionConfig = None,
        gui_id: str | None = None,
        theme_update: bool = False,
        **kwargs,
    ):
        """A widget to display all the signals from a given device, and allow getting
        a fresh reading."""
        super().__init__(
            parent=parent,
            client=client,
            config=config,
            gui_id=gui_id,
            theme_update=theme_update,
            **kwargs,
        )
        self.get_bec_shortcuts()
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)
        self._content = QWidget()
        self._layout.addWidget(self._content)
        self._device = device
        self.device = device

    @SafeSlot()
    def _refresh(self):
        if (dev := self.dev.get(self.device)) is not None:
            dev.read(cached=True)
            dev.read_configuration(cached=True)

    def _add_refresh_button(self):
        button_holder = QWidget()
        button_holder.setLayout(QHBoxLayout())
        button_holder.layout().setAlignment(Qt.AlignmentFlag.AlignRight)
        button_holder.layout().setContentsMargins(0, 0, 0, 0)
        refresh_button = QToolButton()
        refresh_button.setIcon(
            material_icon(icon_name="refresh", size=(20, 20), convert_to_pixmap=False)
        )
        refresh_button.clicked.connect(self._refresh)
        button_holder.layout().addWidget(refresh_button)
        self._content_layout.addWidget(button_holder)

    def _populate(self):
        self._content.deleteLater()
        self._content = QWidget()
        self._layout.addWidget(self._content)
        self._content_layout = QVBoxLayout()
        self._content_layout.setContentsMargins(0, 0, 0, 0)
        self._content.setLayout(self._content_layout)

        self._add_refresh_button()

        if self._device in self.dev:
            if isinstance(self.dev[self.device], Device):
                for sig, info in self.dev[self.device]._info.get("signals", {}).items():
                    if info.get("kind_str") in [
                        Kind.hinted.name,
                        Kind.normal.name,
                        Kind.config.name,
                    ]:
                        self._content_layout.addWidget(
                            SignalLabel(
                                parent=self,
                                device=self._device,
                                signal=sig,
                                show_select_button=False,
                                show_default_units=True,
                            )
                        )
            else:
                self._content_layout.addWidget(
                    SignalLabel(
                        parent=self,
                        device=self._device,
                        signal=self._device,
                        show_select_button=False,
                        show_default_units=True,
                    )
                )
            self._content_layout.addStretch(1)
        else:
            self._content_layout.addWidget(
                QLabel(f"Device {self.device} not found in device manager!")
            )

    @SafeProperty(str)
    def device(self):
        return self._device

    @device.setter
    def device(self, value: str):
        self._device = value
        self._populate()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    from bec_widgets.utils.colors import apply_theme

    app = QApplication(sys.argv)
    apply_theme("light")
    widget = SignalDisplay(device="samx")
    widget.show()
    sys.exit(app.exec_())



from random import randint
from typing import Any, Iterable
from uuid import uuid4

from qtpy.QtCore import QItemSelection, Signal  # type: ignore
from qtpy.QtWidgets import QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.widgets.control.device_manager.components._util import (
    SharedSelectionSignal,
    yield_only_passing,
)
from bec_widgets.widgets.control.device_manager.components.available_device_resources.available_device_resources_ui import (
    Ui_availableDeviceResources,
)
from bec_widgets.widgets.control.device_manager.components.available_device_resources.device_resource_backend import (
    HashableDevice,
    get_backend,
)
from bec_widgets.widgets.control.device_manager.components.constants import CONFIG_DATA_ROLE


class AvailableDeviceResources(BECWidget, QWidget, Ui_availableDeviceResources):

    selected_devices = Signal(list)  # list[dict[str,Any]] of device configs currently selected
    add_selected_devices = Signal(list)
    del_selected_devices = Signal(list)

    def __init__(self, parent=None, shared_selection_signal=SharedSelectionSignal(), **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.setupUi(self)
        self._backend = get_backend()
        self._shared_selection_signal = shared_selection_signal
        self._shared_selection_uuid = str(uuid4())
        self._shared_selection_signal.proc.connect(self._handle_shared_selection_signal)
        self.device_groups_list.selectionModel().selectionChanged.connect(
            self._on_selection_changed
        )
        self.grouping_selector.addItem("deviceTags")
        self.grouping_selector.addItems(self._backend.allowed_sort_keys)
        self._grouping_selection_changed("deviceTags")
        self.grouping_selector.currentTextChanged.connect(self._grouping_selection_changed)
        self.search_box.textChanged.connect(self.device_groups_list.update_filter)

        self.tb_add_selected.action.triggered.connect(self._add_selected_action)
        self.tb_del_selected.action.triggered.connect(self._del_selected_action)

    def refresh_full_list(self, device_groups: dict[str, set[HashableDevice]]):
        self.device_groups_list.clear()
        for device_group, devices in device_groups.items():
            self._add_device_group(device_group, devices)
        if self.grouping_selector.currentText == "deviceTags":
            self._add_device_group("Untagged devices", self._backend.untagged_devices)
        self.device_groups_list.sortItems()

    def _add_device_group(self, device_group: str, devices: set[HashableDevice]):
        item, widget = self.device_groups_list.add_item(
            device_group,
            self.device_groups_list,
            device_group,
            devices,
            shared_selection_signal=self._shared_selection_signal,
            expanded=False,
        )
        item.setData(CONFIG_DATA_ROLE, widget.create_mime_data())
        # Re-emit the selected items from a subgroup - all other selections should be disabled anyway
        widget.selected_devices.connect(self.selected_devices)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        for list_item, device_group_widget in self.device_groups_list.item_widget_pairs():
            list_item.setSizeHint(device_group_widget.sizeHint())

    @SafeSlot()
    def _add_selected_action(self):
        self.add_selected_devices.emit(self.device_groups_list.any_selected_devices())

    @SafeSlot()
    def _del_selected_action(self):
        self.del_selected_devices.emit(self.device_groups_list.any_selected_devices())

    @SafeSlot(QItemSelection, QItemSelection)
    def _on_selection_changed(self, selected: QItemSelection, deselected: QItemSelection) -> None:
        self.selected_devices.emit(self.device_groups_list.selected_devices_from_groups())
        self._shared_selection_signal.proc.emit(self._shared_selection_uuid)

    @SafeSlot(str)
    def _handle_shared_selection_signal(self, uuid: str):
        if uuid != self._shared_selection_uuid:
            self.device_groups_list.clearSelection()

    def _set_devices_state(self, devices: Iterable[HashableDevice], included: bool):
        for device in devices:
            for device_group in self.device_groups_list.widgets():
                device_group.set_item_state(hash(device), included)

    @SafeSlot(list)
    def mark_devices_used(self, config_list: list[dict[str, Any]], used: bool):
        """Set the display color of individual devices and update the group display of numbers
        included. Accepts a list of dicts with the complete config as used in
        bec_lib.atlas_models.Device."""
        self._set_devices_state(
            yield_only_passing(HashableDevice.model_validate, config_list), used
        )

    @SafeSlot(str)
    def _grouping_selection_changed(self, sort_key: str):
        self.search_box.setText("")
        if sort_key == "deviceTags":
            device_groups = self._backend.tag_groups
        else:
            device_groups = self._backend.group_by_key(sort_key)
        self.refresh_full_list(device_groups)


if __name__ == "__main__":
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    widget = AvailableDeviceResources()
    widget._set_devices_state(
        list(filter(lambda _: randint(0, 1) == 1, widget._backend.all_devices)), True
    )
    widget.show()
    sys.exit(app.exec())



from qtpy.QtCore import QTimer
from qtpy.QtGui import QFontMetrics, QPainter
from qtpy.QtWidgets import QLabel


class ScrollLabel(QLabel):
    """A QLabel that scrolls its text horizontally across the widget."""

    def __init__(self, parent=None, speed_ms=30, step_px=1, delay_ms=2000):
        super().__init__(parent=parent)
        self._offset = 0
        self._text_width = 0

        # scrolling timer (runs continuously once started)
        self._timer = QTimer(self)
        self._timer.setInterval(speed_ms)
        self._timer.timeout.connect(self._scroll)

        # delay‑before‑scroll timer (single‑shot)
        self._delay_timer = QTimer(self)
        self._delay_timer.setSingleShot(True)
        self._delay_timer.setInterval(delay_ms)
        self._delay_timer.timeout.connect(self._timer.start)

        self._step_px = step_px

    def setText(self, text):
        """
        Overridden to ensure that new text replaces the current one
        immediately.
        If the label was already scrolling (or in its delay phase),
        the next message starts **without** the extra delay.
        """
        # Determine whether the widget was already in a scrolling cycle
        was_scrolling = self._timer.isActive() or self._delay_timer.isActive()

        super().setText(text)

        fm = QFontMetrics(self.font())
        self._text_width = fm.horizontalAdvance(text)
        self._offset = 0

        # Skip the delay when we were already scrolling
        self._update_timer(skip_delay=was_scrolling)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._update_timer()

    def _update_timer(self, *, skip_delay: bool = False):
        """
        Decide whether to start or stop scrolling.

        If the text is wider than the visible area, start a single‑shot
        delay timer (2s by default). Scrolling begins only after this
        delay. Any change (resize or new text) restarts the logic.
        """
        needs_scroll = self._text_width > self.width()

        if needs_scroll:
            # Reset any running timers
            if self._timer.isActive():
                self._timer.stop()
            if self._delay_timer.isActive():
                self._delay_timer.stop()

            self._offset = 0

            # Start scrolling immediately when we should skip the delay,
            # otherwise apply the configured delay_ms interval
            if skip_delay:
                self._timer.start()
            else:
                self._delay_timer.start()
        else:
            if self._delay_timer.isActive():
                self._delay_timer.stop()
            if self._timer.isActive():
                self._timer.stop()
            self.update()

    def _scroll(self):
        self._offset += self._step_px
        if self._offset >= self._text_width:
            self._offset = 0
        self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.TextAntialiasing)
        text = self.text()
        if not text:
            return
        fm = QFontMetrics(self.font())
        y = (self.height() + fm.ascent() - fm.descent()) // 2
        if self._text_width <= self.width():
            painter.drawText(0, y, text)
        else:
            x = -self._offset
            gap = 50  # space between repeating text blocks
            while x < self.width():
                painter.drawText(x, y, text)
                x += self._text_width + gap

    def cleanup(self):
        """Stop all timers to prevent memory leaks."""
        if self._timer.isActive():
            self._timer.stop()
        if self._delay_timer.isActive():
            self._delay_timer.stop()



from __future__ import annotations

from bec_qthemes import material_icon
from qtpy.QtCore import QSize, Signal
from qtpy.QtWidgets import (
    QApplication,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLayout,
    QSizePolicy,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.clickable_label import ClickableLabel
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot


class ExpandableGroupFrame(QFrame):
    broadcast_size_hint = Signal(QSize)
    imminent_deletion = Signal()
    expansion_state_changed = Signal()

    EXPANDED_ICON_NAME: str = "collapse_all"
    COLLAPSED_ICON_NAME: str = "expand_all"

    def __init__(
        self, parent: QWidget | None = None, title: str = "", expanded: bool = True, icon: str = ""
    ) -> None:
        super().__init__(parent=parent)
        self._expanded = expanded

        self._title_text = f"<b>{title}</b>"
        self.setFrameStyle(QFrame.Shape.StyledPanel | QFrame.Shadow.Raised)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(5, 0, 0, 0)
        self.setLayout(self._layout)

        self._create_title_layout(title, icon)

        self._contents = QWidget(self)
        self._layout.addWidget(self._contents)

        self._expansion_button.clicked.connect(self.switch_expanded_state)
        self.expanded = self._expanded  # type: ignore
        self.expansion_state_changed.emit()

    def _create_title_layout(self, title: str, icon: str):
        self._title_layout = QHBoxLayout()
        self._layout.addLayout(self._title_layout)
        self._internal_title_layout = QHBoxLayout()
        self._title_layout.addLayout(self._internal_title_layout)

        self._title = ClickableLabel()
        self._set_title_text(self._title_text)
        self._title_icon = ClickableLabel()
        self._internal_title_layout.addWidget(self._title_icon)
        self._internal_title_layout.addWidget(self._title)
        self.icon_name = icon
        self._title.clicked.connect(self.switch_expanded_state)
        self._title_icon.clicked.connect(self.switch_expanded_state)

        self._internal_title_layout.addStretch(1)

        self._expansion_button = QToolButton()
        self._update_expansion_icon()
        self._title_layout.addWidget(self._expansion_button, stretch=1)

    def get_title_layout(self) -> QHBoxLayout:
        return self._internal_title_layout

    def set_layout(self, layout: QLayout) -> None:
        self._contents.setLayout(layout)
        self._contents.layout().setContentsMargins(0, 0, 0, 0)  # type: ignore

    @SafeSlot()
    def switch_expanded_state(self):
        self.expanded = not self.expanded  # type: ignore
        self._update_expansion_icon()
        self.expansion_state_changed.emit()

    @SafeProperty(bool)
    def expanded(self):  # type: ignore
        return self._expanded

    @expanded.setter
    def expanded(self, expanded: bool):
        self._expanded = expanded
        self._contents.setVisible(expanded)
        self.updateGeometry()
        self.adjustSize()

    def _update_expansion_icon(self):
        self._expansion_button.setIcon(
            material_icon(icon_name=self.EXPANDED_ICON_NAME, size=(10, 10), convert_to_pixmap=False)
            if self.expanded
            else material_icon(
                icon_name=self.COLLAPSED_ICON_NAME, size=(10, 10), convert_to_pixmap=False
            )
        )

    @SafeProperty(str)
    def icon_name(self):  # type: ignore
        return self._title_icon_name

    @icon_name.setter
    def icon_name(self, icon_name: str):
        self._title_icon_name = icon_name
        self._set_title_icon(self._title_icon_name)

    def _set_title_icon(self, icon_name: str):
        if icon_name:
            self._title_icon.setVisible(True)
            self._title_icon.setPixmap(
                material_icon(icon_name=icon_name, size=(20, 20), convert_to_pixmap=True)
            )
        else:
            self._title_icon.setVisible(False)

    @SafeProperty(str)
    def title_text(self):  # type: ignore
        return self._title_text

    @title_text.setter
    def title_text(self, title_text: str):
        self._title_text = title_text
        self._set_title_text(self._title_text)

    def _set_title_text(self, title_text: str):
        self._title.setText(title_text)


# Application example
if __name__ == "__main__":  # pragma: no cover

    app = QApplication([])
    frame = ExpandableGroupFrame()
    layout = QVBoxLayout()
    frame.set_layout(layout)
    layout.addWidget(QLabel("test1"))
    layout.addWidget(QLabel("test2"))
    layout.addWidget(QLabel("test3"))

    frame.show()
    app.exec()



from bec_qthemes import material_icon
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QHBoxLayout, QPushButton, QToolButton, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeSlot


class ResumeButton(BECWidget, QWidget):
    """A button that continue scan queue."""

    PLUGIN = True
    ICON_NAME = "resume"
    RPC = True

    def __init__(self, parent=None, client=None, config=None, gui_id=None, toolbar=False, **kwargs):
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        self.get_bec_shortcuts()

        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        if toolbar:
            icon = material_icon("resume", color="#2793e8", filled=True, convert_to_pixmap=False)
            self.button = QToolButton(icon=icon)
            self.button.setToolTip("Resume the scan queue")
        else:
            self.button = QPushButton()
            self.button.setText("Resume")
            self.button.setStyleSheet(
                "background-color:  #2793e8; color: white; font-weight: bold; font-size: 12px;"
            )
        self.button.clicked.connect(self.continue_scan)

        self.layout.addWidget(self.button)

    @SafeSlot()
    def continue_scan(self):
        """Stop the scan."""
        self.queue.request_scan_continuation()



from .scan_history_device_viewer import ScanHistoryDeviceViewer
from .scan_history_metadata_viewer import ScanHistoryMetadataViewer
from .scan_history_view import ScanHistoryView

__all__ = ["ScanHistoryDeviceViewer", "ScanHistoryMetadataViewer", "ScanHistoryView"]
